import unittest

from test.testutil import set_default_576_324_videos_for_testing
from vmaf.config import VmafConfig
from vmaf.core.quality_runner import PsnrQualityRunner
from vmaf.tools.misc import MyTestCase, QualityRunnerTestMixin, import_python_file


class QualityRunnerTestMixinTest(MyTestCase, QualityRunnerTestMixin):

    def setUp(self):
        super().setUp()
        ref_path, dis_path, asset, asset_original = set_default_576_324_videos_for_testing()
        self.data = (
            [30.755063979166664, PsnrQualityRunner, asset, None],
        )

    def test_run_each(self):
        for data_each in self.data:
            self.run_each(*data_each)

    def test_plot_frame_scores(self):
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        for data_each in self.data:
            new_data_each = data_each[1:]
            self.plot_frame_scores(ax, *new_data_each, label='src01_hrc01_576x324.yuv')
        ax.legend()
        ax.grid()
        # from vmaf.config import DisplayConfig; DisplayConfig.show(write_to_dir=None)


class MiscTest(unittest.TestCase):

    def test_import_python_file(self):
        ds = import_python_file(VmafConfig.test_resource_path('example_raw_dataset2.py'))
        self.assertEqual(ds.dataset_name, 'example')
        self.assertEqual(ds.yuv_fmt, 'yuv420p')
        self.assertEqual(ds.ref_videos[0]['content_id'], 0)
        self.assertEqual(ds.ref_videos[0]['path'], '/dir/yuv/checkerboard_1920_1080_10_3_0_0.yuv')
        self.assertEqual(ds.dis_videos[3]['content_id'], 1)
        self.assertEqual(ds.dis_videos[3]['asset_id'], 3)
        self.assertEqual(ds.dis_videos[3]['os'], [70, 75, 80, 85, 90])
        self.assertEqual(ds.dis_videos[3]['path'], '/dir/yuv/flat_1920_1080_10.yuv')

    def test_import_python_file_with_override(self):
        ds = import_python_file(VmafConfig.test_resource_path('example_raw_dataset2.py'), override={'dir': '/xyz'})
        self.assertEqual(ds.dataset_name, 'example')
        self.assertEqual(ds.yuv_fmt, 'yuv420p')
        self.assertEqual(ds.ref_videos[0]['content_id'], 0)
        self.assertEqual(ds.ref_videos[0]['path'], '/xyz/yuv/checkerboard_1920_1080_10_3_0_0.yuv')
        self.assertEqual(ds.dis_videos[3]['content_id'], 1)
        self.assertEqual(ds.dis_videos[3]['asset_id'], 3)
        self.assertEqual(ds.dis_videos[3]['os'], [70, 75, 80, 85, 90])
        self.assertEqual(ds.dis_videos[3]['path'], '/xyz/yuv/flat_1920_1080_10.yuv')

    def test_import_python_file_with_override_and_add_new(self):
        ds = import_python_file(VmafConfig.test_resource_path('example_raw_dataset3.py'),
                                override={'dir': '/xyz', 'quality_width': 960, 'quality_height': 540})
        self.assertEqual(ds.dataset_name, 'example')
        self.assertEqual(ds.yuv_fmt, 'yuv420p')
        self.assertEqual(ds.ref_videos[0]['content_id'], 0)
        self.assertEqual(ds.ref_videos[0]['path'], '/xyz/yuv/checkerboard_1920_1080_10_3_0_0.yuv')
        self.assertEqual(ds.dis_videos[3]['content_id'], 1)
        self.assertEqual(ds.dis_videos[3]['asset_id'], 3)
        self.assertEqual(ds.dis_videos[3]['os'], [70, 75, 80, 85, 90])
        self.assertEqual(ds.dis_videos[3]['path'], '/xyz/yuv/flat_1920_1080_10.yuv')
        self.assertEqual(ds.quality_width, 960)
        self.assertEqual(ds.quality_height, 540)

    def test_import_python_file_with_override_multiple_in_one_line(self):
        ds = import_python_file(VmafConfig.test_resource_path('example_raw_dataset4.py'),
                                override={'dir': '/xyz', 'quality_width': 960, 'quality_height': 540})
        self.assertEqual(ds.dataset_name, 'example')
        self.assertEqual(ds.yuv_fmt, 'yuv420p')
        self.assertEqual(ds.ref_videos[0]['content_id'], 0)
        self.assertEqual(ds.ref_videos[0]['path'], '/xyz/yuv/checkerboard_1920_1080_10_3_0_0.yuv')
        self.assertEqual(ds.dis_videos[3]['content_id'], 1)
        self.assertEqual(ds.dis_videos[3]['asset_id'], 3)
        self.assertEqual(ds.dis_videos[3]['os'], [70, 75, 80, 85, 90])
        self.assertEqual(ds.dis_videos[3]['path'], '/xyz/yuv/flat_1920_1080_10.yuv')
        self.assertEqual(ds.quality_width, 960)
        self.assertEqual(ds.quality_height, 540)




from abc import ABCMeta, abstractmethod
from xml.etree import ElementTree

from vmaf.tools.decorator import override

__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"

import re
import numpy as np
import ast

from vmaf import ExternalProgramCaller
from vmaf.core.executor import Executor
from vmaf.core.result import Result, BasicResult
from vmaf.tools.reader import YuvReader


class FeatureExtractor(Executor):
    """
    FeatureExtractor takes in a list of assets, and run feature extraction on
    them, and return a list of corresponding results. A FeatureExtractor must
    specify a unique type and version combination (by the TYPE and VERSION
    attribute), so that the Result generated by it can be identified.

    A derived class of FeatureExtractor must:
        1) Override TYPE and VERSION
        2) Override _generate_result(self, asset), which call a
        command-line executable and generate feature scores in a log file.
        3) Override _get_feature_scores(self, asset), which read the feature
        scores from the log file, and return the scores in a dictionary format.
    For an example, follow VmafFeatureExtractor.
    """

    __metaclass__ = ABCMeta

    @property
    @abstractmethod
    def ATOM_FEATURES(self):
        raise NotImplementedError

    def _read_result(self, asset):
        result = {}
        result.update(self._get_feature_scores(asset))
        executor_id = self.executor_id
        return Result(asset, executor_id, result)

    @classmethod
    def get_scores_key(cls, atom_feature):
        return "{type}_{atom_feature}_scores".format(
            type=cls.TYPE, atom_feature=atom_feature)

    @classmethod
    def get_score_key(cls, atom_feature):
        return "{type}_{atom_feature}_score".format(
            type=cls.TYPE, atom_feature=atom_feature)

    def _get_feature_scores(self, asset):
        # routine to read the feature scores from the log file, and return
        # the scores in a dictionary format.

        log_file_path = self._get_log_file_path(asset)

        atom_feature_scores_dict = {}
        atom_feature_idx_dict = {}
        for atom_feature in self.ATOM_FEATURES:
            atom_feature_scores_dict[atom_feature] = []
            atom_feature_idx_dict[atom_feature] = 0

        with open(log_file_path, 'rt') as log_file:
            for line in log_file.readlines():
                for atom_feature in self.ATOM_FEATURES:
                    re_template = "{af}: ([0-9]+) ([a-zA-Z0-9.-]+)".format(af=atom_feature)
                    mo = re.match(re_template, line)
                    if mo:

                        cur_idx = int(mo.group(1))
                        assert cur_idx == atom_feature_idx_dict[atom_feature]

                        # parse value, allowing NaN and inf
                        val = float(mo.group(2))
                        if np.isnan(val) or np.isinf(val):
                            val = None

                        atom_feature_scores_dict[atom_feature].append(val)
                        atom_feature_idx_dict[atom_feature] += 1
                        continue

        len_score = len(atom_feature_scores_dict[self.ATOM_FEATURES[0]])
        assert len_score != 0
        for atom_feature in self.ATOM_FEATURES[1:]:
            assert len_score == len(atom_feature_scores_dict[atom_feature]), \
                "Feature data possibly corrupt. Run cleanup script and try again."

        feature_result = {}

        for atom_feature in self.ATOM_FEATURES:
            scores_key = self.get_scores_key(atom_feature)
            feature_result[scores_key] = atom_feature_scores_dict[atom_feature]

        return feature_result


class VmafexecFeatureExtractorMixin(object):

    @override(FeatureExtractor)
    def _get_feature_scores(self, asset):

        assert hasattr(self, '_get_log_file_path')
        assert hasattr(self, 'ATOM_FEATURES')
        assert hasattr(self, 'ATOM_FEATURES_TO_VMAFEXEC_KEY_DICT')
        assert hasattr(self, 'get_scores_key')

        log_file_path = self._get_log_file_path(asset)
        tree = ElementTree.parse(log_file_path)
        root = tree.getroot()

        feature_scores = [[] for _ in self.ATOM_FEATURES]
        feature_nicknames = [None for _ in self.ATOM_FEATURES]

        for frame in root.findall('frames/frame'):
            for i_feature, feature in enumerate(self.ATOM_FEATURES):

                # first look for exact match
                feature_found = self._discover_feature_exact(
                    frame, i_feature, feature, feature_scores, feature_nicknames)

                if feature_found:
                    continue

                # wildcard discovery: look for xxx features
                feature_found = self._discover_feature_wildcard(
                    frame, i_feature, feature, feature_scores, feature_nicknames)

        for i_feature, feature in enumerate(self.ATOM_FEATURES):
           if len(feature_scores[i_feature]) != 0:
               assert len(feature_scores[i_feature]) == len(feature_scores[0])

        feature_result = {}
        for i_feature, feature in enumerate(self.ATOM_FEATURES):
            if len(feature_scores[i_feature]) != 0:
                assert feature_nicknames[i_feature] is not None
                feature_result[self.get_scores_key(feature_nicknames[i_feature])] = feature_scores[i_feature]

        return feature_result

    @classmethod
    def _discover_feature_exact(cls, frame, i_feature, feature,
                                feature_scores, feature_nicknames):
        assert hasattr(cls, 'ATOM_FEATURES_TO_VMAFEXEC_KEY_DICT')
        feature_found = False
        for feature_fullname in frame.attrib:
            if cls.ATOM_FEATURES_TO_VMAFEXEC_KEY_DICT[feature] == feature_fullname:
                feature_scores[i_feature].append(
                    float(frame.attrib[feature_fullname]))
                if feature_nicknames[i_feature] is None:
                    feature_nicknames[i_feature] = feature
                else:
                    assert feature_nicknames[i_feature] == feature
                feature_found = True
                break
        return feature_found

    @classmethod
    def _discover_feature_wildcard(cls, frame, i_feature, feature,
                                   feature_scores, feature_nicknames):
        assert hasattr(cls, 'ATOM_FEATURES_TO_VMAFEXEC_KEY_DICT')
        feature_prefix = cls.ATOM_FEATURES_TO_VMAFEXEC_KEY_DICT[feature] + '_'
        feature_found = False
        for feature_fullname in frame.attrib:
            if feature_fullname.startswith(feature_prefix):
                feature_scores[i_feature].append(
                    float(frame.attrib[feature_fullname]))
                feature_suffix = feature_fullname[len(feature_prefix):]
                feature_nickname = feature + '_' + feature_suffix
                if feature_nicknames[i_feature] is None:
                    feature_nicknames[i_feature] = feature_nickname
                else:
                    assert feature_nicknames[i_feature] == feature_nickname
                feature_found = True
                break
        return feature_found


class VmafFeatureExtractor(VmafexecFeatureExtractorMixin, FeatureExtractor):

    TYPE = "VMAF_feature"

    # VERSION = '0.1' # vmaf_study; Anush's VIF fix
    # VERSION = '0.2' # expose vif_num, vif_den, adm_num, adm_den, anpsnr
    # VERSION = '0.2.1' # expose vif num/den of each scale
    # VERSION = '0.2.2'  # adm abs-->fabs, corrected border handling, uniform reading with option of offset for input YUV, updated VIF corner case
    # VERSION = '0.2.2b'  # expose adm_den/num_scalex
    # VERSION = '0.2.3'  # AVX for VMAF convolution; update adm features by folding noise floor into per coef
    # VERSION = '0.2.4'  # Fix a bug in adm feature passing scale into dwt_quant_step
    # VERSION = '0.2.4b'  # Modify by adding ADM noise floor outside cube root; add derived feature motion2
    # VERSION = '0.2.4c'  # Modify by moving motion2 to c code
    # VERSION = '0.2.5'  # replace executable vmaf_feature with vmaf
    # VERSION = '0.2.6'  # incorporate adm_enhn_gain_limit and vif_enhn_gain_limit
    VERSION = '0.2.7'  # move vif_enhn_gain_limit right before log calculation

    ATOM_FEATURES = ['vif', 'adm', 'ansnr', 'motion', 'motion2',
                     'vif_num', 'vif_den', 'adm_num', 'adm_den', 'anpsnr',
                     'vif_num_scale0', 'vif_den_scale0',
                     'vif_num_scale1', 'vif_den_scale1',
                     'vif_num_scale2', 'vif_den_scale2',
                     'vif_num_scale3', 'vif_den_scale3',
                     'adm_num_scale0', 'adm_den_scale0',
                     'adm_num_scale1', 'adm_den_scale1',
                     'adm_num_scale2', 'adm_den_scale2',
                     'adm_num_scale3', 'adm_den_scale3',
                     ]

    ATOM_FEATURES_TO_VMAFEXEC_KEY_DICT = dict(zip(ATOM_FEATURES, ATOM_FEATURES))
    ATOM_FEATURES_TO_VMAFEXEC_KEY_DICT['ansnr'] = 'float_ansnr'
    ATOM_FEATURES_TO_VMAFEXEC_KEY_DICT['anpsnr'] = 'float_anpsnr'

    DERIVED_ATOM_FEATURES = ['vif_scale0', 'vif_scale1', 'vif_scale2', 'vif_scale3',
                             'vif2', 'adm2', 'adm3',
                             'adm_scale0', 'adm_scale1', 'adm_scale2', 'adm_scale3',
                             ]

    ADM2_CONSTANT = 0
    ADM_SCALE_CONSTANT = 0

    def _generate_result(self, asset):
        # routine to call the command-line executable and generate feature
        # scores in the log file.

        quality_width, quality_height = asset.quality_width_height
        log_file_path = self._get_log_file_path(asset)

        yuv_type = self._get_workfile_yuv_type(asset)
        ref_path = asset.ref_procfile_path
        dis_path = asset.dis_procfile_path
        w = quality_width
        h = quality_height
        logger = self.logger

        features = ['float_adm', 'float_vif', 'float_motion', 'float_ansnr']
        options = {
            'float_adm': {'debug': True},
            'float_vif': {'debug': True},
            'float_motion': {'debug': True},
        }

        if self.optional_dict is not None:
            for opt in self.optional_dict:
                if opt == 'vif_enhn_gain_limit':
                    options['float_vif']['vif_enhn_gain_limit'] = self.optional_dict['vif_enhn_gain_limit']
                elif opt == 'vif_kernelscale':
                    options['float_vif']['vif_kernelscale'] = self.optional_dict['vif_kernelscale']
                elif opt == 'adm_csf_mode':
                    options['float_adm']['adm_csf_mode'] = self.optional_dict['adm_csf_mode']
                elif opt == 'adm_enhn_gain_limit':
                    options['float_adm']['adm_enhn_gain_limit'] = self.optional_dict['adm_enhn_gain_limit']
                elif opt == 'adm_norm_view_dist':
                    options['float_adm']['adm_norm_view_dist'] = self.optional_dict['adm_norm_view_dist']
                elif opt == 'adm_ref_display_height':
                    options['float_adm']['adm_ref_display_height'] = self.optional_dict['adm_ref_display_height']
                elif opt == 'motion_force_zero':
                    options['float_motion']['motion_force_zero'] = self.optional_dict['motion_force_zero']
                else:
                    pass

        if self.optional_dict2 is not None:
            for opt in self.optional_dict2:
                if opt == 'disable_avx':
                    options['disable_avx'] = self.optional_dict2['disable_avx']
                elif opt == 'n_threads':
                    options['n_threads'] = self.optional_dict2['n_threads']

        ExternalProgramCaller.call_vmafexec_multi_features(
            features, yuv_type, ref_path, dis_path, w, h, log_file_path,
            logger, options=options
        )

    @classmethod
    @override(Executor)
    def _post_process_result(cls, result):

        result = super(VmafFeatureExtractor, cls)._post_process_result(result)

        # adm2 =
        # (adm_num + ADM2_CONSTANT) / (adm_den + ADM2_CONSTANT)
        adm2_scores_key = cls.get_scores_key('adm2')
        adm_num_scores_key = BasicResult.scores_key_wildcard_match(result.result_dict, cls.get_scores_key('adm_num'))
        adm_den_scores_key = BasicResult.scores_key_wildcard_match(result.result_dict, cls.get_scores_key('adm_den'))
        result.result_dict[adm2_scores_key] = list(
            (np.array(result.result_dict[adm_num_scores_key]) + cls.ADM2_CONSTANT) /
            (np.array(result.result_dict[adm_den_scores_key]) + cls.ADM2_CONSTANT)
        )

        # vif_scalei = vif_num_scalei / vif_den_scalei, i = 0, 1, 2, 3
        vif_num_scale0_scores_key = BasicResult.scores_key_wildcard_match(result.result_dict, cls.get_scores_key('vif_num_scale0'))
        vif_den_scale0_scores_key = BasicResult.scores_key_wildcard_match(result.result_dict, cls.get_scores_key('vif_den_scale0'))
        vif_num_scale1_scores_key = BasicResult.scores_key_wildcard_match(result.result_dict, cls.get_scores_key('vif_num_scale1'))
        vif_den_scale1_scores_key = BasicResult.scores_key_wildcard_match(result.result_dict, cls.get_scores_key('vif_den_scale1'))
        vif_num_scale2_scores_key = BasicResult.scores_key_wildcard_match(result.result_dict, cls.get_scores_key('vif_num_scale2'))
        vif_den_scale2_scores_key = BasicResult.scores_key_wildcard_match(result.result_dict, cls.get_scores_key('vif_den_scale2'))
        vif_num_scale3_scores_key = BasicResult.scores_key_wildcard_match(result.result_dict, cls.get_scores_key('vif_num_scale3'))
        vif_den_scale3_scores_key = BasicResult.scores_key_wildcard_match(result.result_dict, cls.get_scores_key('vif_den_scale3'))
        vif_scale0_scores_key = cls.get_scores_key('vif_scale0')
        vif_scale1_scores_key = cls.get_scores_key('vif_scale1')
        vif_scale2_scores_key = cls.get_scores_key('vif_scale2')
        vif_scale3_scores_key = cls.get_scores_key('vif_scale3')
        result.result_dict[vif_scale0_scores_key] = list(
            (np.array(result.result_dict[vif_num_scale0_scores_key])
             / np.array(result.result_dict[vif_den_scale0_scores_key]))
        )
        result.result_dict[vif_scale1_scores_key] = list(
            (np.array(result.result_dict[vif_num_scale1_scores_key])
             / np.array(result.result_dict[vif_den_scale1_scores_key]))
        )
        result.result_dict[vif_scale2_scores_key] = list(
            (np.array(result.result_dict[vif_num_scale2_scores_key])
             / np.array(result.result_dict[vif_den_scale2_scores_key]))
        )
        result.result_dict[vif_scale3_scores_key] = list(
            (np.array(result.result_dict[vif_num_scale3_scores_key])
             / np.array(result.result_dict[vif_den_scale3_scores_key]))
        )

        # vif2 =
        # ((vif_num_scale0 / vif_den_scale0) + (vif_num_scale1 / vif_den_scale1) +
        # (vif_num_scale2 / vif_den_scale2) + (vif_num_scale3 / vif_den_scale3)) / 4.0
        vif_scores_key = cls.get_scores_key('vif2')
        result.result_dict[vif_scores_key] = list(
            (
                (np.array(result.result_dict[vif_num_scale0_scores_key])
                 / np.array(result.result_dict[vif_den_scale0_scores_key])) +
                (np.array(result.result_dict[vif_num_scale1_scores_key])
                 / np.array(result.result_dict[vif_den_scale1_scores_key])) +
                (np.array(result.result_dict[vif_num_scale2_scores_key])
                 / np.array(result.result_dict[vif_den_scale2_scores_key])) +
                (np.array(result.result_dict[vif_num_scale3_scores_key])
                 / np.array(result.result_dict[vif_den_scale3_scores_key]))
            ) / 4.0
        )

        # adm_scalei = adm_num_scalei / adm_den_scalei, i = 0, 1, 2, 3
        adm_num_scale0_scores_key = BasicResult.scores_key_wildcard_match(result.result_dict, cls.get_scores_key('adm_num_scale0'))
        adm_den_scale0_scores_key = BasicResult.scores_key_wildcard_match(result.result_dict, cls.get_scores_key('adm_den_scale0'))
        adm_num_scale1_scores_key = BasicResult.scores_key_wildcard_match(result.result_dict, cls.get_scores_key('adm_num_scale1'))
        adm_den_scale1_scores_key = BasicResult.scores_key_wildcard_match(result.result_dict, cls.get_scores_key('adm_den_scale1'))
        adm_num_scale2_scores_key = BasicResult.scores_key_wildcard_match(result.result_dict, cls.get_scores_key('adm_num_scale2'))
        adm_den_scale2_scores_key = BasicResult.scores_key_wildcard_match(result.result_dict, cls.get_scores_key('adm_den_scale2'))
        adm_num_scale3_scores_key = BasicResult.scores_key_wildcard_match(result.result_dict, cls.get_scores_key('adm_num_scale3'))
        adm_den_scale3_scores_key = BasicResult.scores_key_wildcard_match(result.result_dict, cls.get_scores_key('adm_den_scale3'))
        adm_scale0_scores_key = cls.get_scores_key('adm_scale0')
        adm_scale1_scores_key = cls.get_scores_key('adm_scale1')
        adm_scale2_scores_key = cls.get_scores_key('adm_scale2')
        adm_scale3_scores_key = cls.get_scores_key('adm_scale3')
        result.result_dict[adm_scale0_scores_key] = list(
            (np.array(result.result_dict[adm_num_scale0_scores_key]) + cls.ADM_SCALE_CONSTANT)
            / (np.array(result.result_dict[adm_den_scale0_scores_key]) + cls.ADM_SCALE_CONSTANT)
        )
        result.result_dict[adm_scale1_scores_key] = list(
            (np.array(result.result_dict[adm_num_scale1_scores_key]) + cls.ADM_SCALE_CONSTANT)
            / (np.array(result.result_dict[adm_den_scale1_scores_key]) + cls.ADM_SCALE_CONSTANT)
        )
        result.result_dict[adm_scale2_scores_key] = list(
            (np.array(result.result_dict[adm_num_scale2_scores_key]) + cls.ADM_SCALE_CONSTANT)
            / (np.array(result.result_dict[adm_den_scale2_scores_key]) + cls.ADM_SCALE_CONSTANT)
        )
        result.result_dict[adm_scale3_scores_key] = list(
            (np.array(result.result_dict[adm_num_scale3_scores_key]) + cls.ADM_SCALE_CONSTANT)
            / (np.array(result.result_dict[adm_den_scale3_scores_key]) + cls.ADM_SCALE_CONSTANT)
        )

        # adm3 = \
        # (((adm_num_scale0 + ADM_SCALE_CONSTANT) / (adm_den_scale0 + ADM_SCALE_CONSTANT))
        #  + ((adm_num_scale1 + ADM_SCALE_CONSTANT) / (adm_den_scale1 + ADM_SCALE_CONSTANT))
        #  + ((adm_num_scale2 + ADM_SCALE_CONSTANT) / (adm_den_scale2 + ADM_SCALE_CONSTANT))
        #  + ((adm_num_scale3 + ADM_SCALE_CONSTANT) / (adm_den_scale3 + ADM_SCALE_CONSTANT))) / 4.0
        adm3_scores_key = cls.get_scores_key('adm3')
        result.result_dict[adm3_scores_key] = list(
            (
                ((np.array(result.result_dict[adm_num_scale0_scores_key]) + cls.ADM_SCALE_CONSTANT)
                 / (np.array(result.result_dict[adm_den_scale0_scores_key]) + cls.ADM_SCALE_CONSTANT)) +
                ((np.array(result.result_dict[adm_num_scale1_scores_key]) + cls.ADM_SCALE_CONSTANT)
                 / (np.array(result.result_dict[adm_den_scale1_scores_key]) + cls.ADM_SCALE_CONSTANT)) +
                ((np.array(result.result_dict[adm_num_scale2_scores_key]) + cls.ADM_SCALE_CONSTANT)
                 / (np.array(result.result_dict[adm_den_scale2_scores_key]) + cls.ADM_SCALE_CONSTANT)) +
                ((np.array(result.result_dict[adm_num_scale3_scores_key]) + cls.ADM_SCALE_CONSTANT)
                 / (np.array(result.result_dict[adm_den_scale3_scores_key]) + cls.ADM_SCALE_CONSTANT))
            ) / 4.0
        )

        # validate
        for feature in cls.DERIVED_ATOM_FEATURES:
            assert cls.get_scores_key(feature) in result.result_dict

        return result


class VmafIntegerFeatureExtractor(VmafFeatureExtractor):

    TYPE = "VMAF_integer_feature"

    ATOM_FEATURES_TO_VMAFEXEC_KEY_DICT = dict(
        zip(VmafFeatureExtractor.ATOM_FEATURES,
            [f'integer_{f}' for f in VmafFeatureExtractor.ATOM_FEATURES]))
    ATOM_FEATURES_TO_VMAFEXEC_KEY_DICT['ansnr'] = 'float_ansnr'
    ATOM_FEATURES_TO_VMAFEXEC_KEY_DICT['anpsnr'] = 'float_anpsnr'

    def _generate_result(self, asset):

        quality_width, quality_height = asset.quality_width_height
        log_file_path = self._get_log_file_path(asset)

        yuv_type = self._get_workfile_yuv_type(asset)
        ref_path = asset.ref_procfile_path
        dis_path = asset.dis_procfile_path
        w = quality_width
        h = quality_height
        logger = self.logger

        features = ['adm', 'vif', 'motion', 'float_ansnr']
        options = {
            'adm': {'debug': True},
            'vif': {'debug': True},
            'motion': {'debug': True},
        }

        if self.optional_dict is not None:
            for opt in self.optional_dict:
                if opt == 'vif_enhn_gain_limit':
                    options['vif']['vif_enhn_gain_limit'] = self.optional_dict['vif_enhn_gain_limit']
                elif opt == 'adm_enhn_gain_limit':
                    options['adm']['adm_enhn_gain_limit'] = self.optional_dict['adm_enhn_gain_limit']
                elif opt == 'adm_norm_view_dist':
                    options['adm']['adm_norm_view_dist'] = self.optional_dict['adm_norm_view_dist']
                elif opt == 'adm_ref_display_height':
                    options['adm']['adm_ref_display_height'] = self.optional_dict['adm_ref_display_height']
                elif opt == 'motion_force_zero':
                    options['motion']['motion_force_zero'] = self.optional_dict['motion_force_zero']
                else:
                    pass

        if self.optional_dict2 is not None:
            for opt in self.optional_dict2:
                if opt == 'disable_avx':
                    options['disable_avx'] = self.optional_dict2['disable_avx']
                elif opt == 'n_threads':
                    options['n_threads'] = self.optional_dict2['n_threads']

        ExternalProgramCaller.call_vmafexec_multi_features(
            features, yuv_type, ref_path, dis_path, w, h, log_file_path,
            logger, options=options
        )


class VifFrameDifferenceFeatureExtractor(FeatureExtractor):

    TYPE = "VifDiff_feature"

    VERSION = '0.1'

    ATOM_FEATURES = ['vifdiff',
                     'vifdiff_num', 'vifdiff_den',
                     'vifdiff_num_scale0', 'vifdiff_den_scale0',
                     'vifdiff_num_scale1', 'vifdiff_den_scale1',
                     'vifdiff_num_scale2', 'vifdiff_den_scale2',
                     'vifdiff_num_scale3', 'vifdiff_den_scale3',
                     ]

    DERIVED_ATOM_FEATURES = ['vifdiff_scale0', 'vifdiff_scale1', 'vifdiff_scale2', 'vifdiff_scale3',
                             ]

    ADM2_CONSTANT = 0
    ADM_SCALE_CONSTANT = 0

    def _generate_result(self, asset):
        # routine to call the command-line executable and generate feature
        # scores in the log file.

        quality_width, quality_height = asset.quality_width_height
        log_file_path = self._get_log_file_path(asset)

        yuv_type = self._get_workfile_yuv_type(asset)
        ref_path = asset.ref_procfile_path
        dis_path = asset.dis_procfile_path
        w = quality_width
        h = quality_height
        logger = self.logger

        ExternalProgramCaller.call_vifdiff_feature(yuv_type, ref_path, dis_path, w, h, log_file_path, logger)

    @classmethod
    @override(Executor)
    def _post_process_result(cls, result):

        result = super(VifFrameDifferenceFeatureExtractor, cls)._post_process_result(result)

        # vifdiff_scalei = vifdiff_num_scalei / vifdiff_den_scalei, i = 0, 1, 2, 3
        vifdiff_num_scale0_scores_key = cls.get_scores_key('vifdiff_num_scale0')
        vifdiff_den_scale0_scores_key = cls.get_scores_key('vifdiff_den_scale0')
        vifdiff_num_scale1_scores_key = cls.get_scores_key('vifdiff_num_scale1')
        vifdiff_den_scale1_scores_key = cls.get_scores_key('vifdiff_den_scale1')
        vifdiff_num_scale2_scores_key = cls.get_scores_key('vifdiff_num_scale2')
        vifdiff_den_scale2_scores_key = cls.get_scores_key('vifdiff_den_scale2')
        vifdiff_num_scale3_scores_key = cls.get_scores_key('vifdiff_num_scale3')
        vifdiff_den_scale3_scores_key = cls.get_scores_key('vifdiff_den_scale3')
        vifdiff_scale0_scores_key = cls.get_scores_key('vifdiff_scale0')
        vifdiff_scale1_scores_key = cls.get_scores_key('vifdiff_scale1')
        vifdiff_scale2_scores_key = cls.get_scores_key('vifdiff_scale2')
        vifdiff_scale3_scores_key = cls.get_scores_key('vifdiff_scale3')
        result.result_dict[vifdiff_scale0_scores_key] = list(
            (np.array(result.result_dict[vifdiff_num_scale0_scores_key])
             / np.array(result.result_dict[vifdiff_den_scale0_scores_key]))
        )
        result.result_dict[vifdiff_scale1_scores_key] = list(
            (np.array(result.result_dict[vifdiff_num_scale1_scores_key])
             / np.array(result.result_dict[vifdiff_den_scale1_scores_key]))
        )
        result.result_dict[vifdiff_scale2_scores_key] = list(
            (np.array(result.result_dict[vifdiff_num_scale2_scores_key])
             / np.array(result.result_dict[vifdiff_den_scale2_scores_key]))
        )
        result.result_dict[vifdiff_scale3_scores_key] = list(
            (np.array(result.result_dict[vifdiff_num_scale3_scores_key])
             / np.array(result.result_dict[vifdiff_den_scale3_scores_key]))
        )

        # validate
        for feature in cls.DERIVED_ATOM_FEATURES:
            assert cls.get_scores_key(feature) in result.result_dict

        return result


class PyFeatureExtractorMixin(object):

    @override(FeatureExtractor)
    def _get_feature_scores(self, asset):
        """
        Reads the feature scores from a generated log. Assumes that the log contains the per-frame features, e.g.,
        {
            'frame': frm,
            'feature_1': feature_1,
            'feature_2': feature_2,
        }
        """
        log_file_path = self._get_log_file_path(asset)

        with open(log_file_path, 'rt') as log_file:
            log_str = log_file.read()
            log_dicts = ast.literal_eval(log_str)

        feature_result = dict()
        frm = 0
        for log_dict in log_dicts:
            assert frm == log_dict['frame']
            for feature in self.ATOM_FEATURES:
                feature_result.setdefault(self.get_scores_key(feature), []).append(log_dict[feature])
            frm += 1

        return feature_result


class PypsnrFeatureExtractor(PyFeatureExtractorMixin, FeatureExtractor):

    TYPE = "Pypsnr_feature"
    VERSION = "1.0"

    ATOM_FEATURES = ['psnry', 'psnru', 'psnrv']

    @staticmethod
    def _assert_bit_depth(ref_yuv_reader, dis_yuv_reader):
        if ref_yuv_reader._is_8bit():
            assert dis_yuv_reader._is_8bit()
        elif ref_yuv_reader._is_10bitle():
            assert dis_yuv_reader._is_10bitle()
        elif ref_yuv_reader._is_12bitle():
            assert dis_yuv_reader._is_12bitle()
        elif ref_yuv_reader._is_16bitle():
            assert dis_yuv_reader._is_16bitle()
        else:
            assert False, 'unknown bit depth and type'

    def _get_max_db(self, ref_yuv_reader):
        if self.optional_dict is not None and 'max_db' in self.optional_dict:
            max_db = self.optional_dict['max_db']
            assert isinstance(max_db, (int, float))
            return max_db
        elif ref_yuv_reader._is_8bit():
            return 60.0
        elif ref_yuv_reader._is_10bitle():
            return 72.0
        elif ref_yuv_reader._is_12bitle():
            return 84.0
        elif ref_yuv_reader._is_16bitle():
            return 108.0
        else:
            assert False, 'unknown bit depth and type'

    def _generate_result(self, asset):
        quality_w, quality_h = asset.quality_width_height
        yuv_type = self._get_workfile_yuv_type(asset)
        log_dicts = list()
        with YuvReader(filepath=asset.ref_procfile_path, width=quality_w, height=quality_h,
                       yuv_type=yuv_type) as ref_yuv_reader:
            with YuvReader(filepath=asset.dis_procfile_path, width=quality_w, height=quality_h,
                           yuv_type=yuv_type) as dis_yuv_reader:

                self._assert_bit_depth(ref_yuv_reader, dis_yuv_reader)
                max_db = self._get_max_db(ref_yuv_reader)

                frm = 0
                while True:
                    try:
                        ref_yuv = ref_yuv_reader.next(format='float')
                        dis_yuv = dis_yuv_reader.next(format='float')
                    except StopIteration:
                        break

                    ref_y, ref_u, ref_v = ref_yuv
                    dis_y, dis_u, dis_v = dis_yuv
                    mse_y, mse_u, mse_v = (
                        np.mean((ref_y - dis_y)**2) + 1e-16,
                        np.mean((ref_u - dis_u)**2) + 1e-16,
                        np.mean((ref_v - dis_v)**2) + 1e-16
                    )
                    psnr_y, psnr_u, psnr_v = (
                        min(10 * np.log10(1.0 / mse_y), max_db),
                        min(10 * np.log10(1.0 / mse_u), max_db),
                        min(10 * np.log10(1.0 / mse_v), max_db)
                    )

                    log_dicts.append({
                        'frame': frm,
                        'psnry': psnr_y,
                        'psnru': psnr_u,
                        'psnrv': psnr_v,
                    })

                    frm += 1

        log_file_path = self._get_log_file_path(asset)
        with open(log_file_path, 'wt') as log_file:
            log_file.write(str(log_dicts))


class PypsnrMaxdb100FeatureExtractor(PypsnrFeatureExtractor):

    TYPE = "Pypsnr_maxdb100_feature"

    @override(Executor)
    def _custom_init(self):
        super()._custom_init()
        if self.optional_dict is not None:
            assert 'max_db' not in self.optional_dict
        if self.optional_dict is None:
            self.optional_dict = dict()
        self.optional_dict['max_db'] = 100.0


class PsnrFeatureExtractor(VmafexecFeatureExtractorMixin, FeatureExtractor):

    TYPE = "PSNR_feature"
    # VERSION = "1.0"
    VERSION = "1.1"  # call vmaf to replace standalone psnr exec

    ATOM_FEATURES = ['psnr']

    ATOM_FEATURES_TO_VMAFEXEC_KEY_DICT = {
        'psnr': 'float_psnr',
    }

    def _generate_result(self, asset):
        # routine to call the command-line executable and generate quality
        # scores in the log file.

        quality_width, quality_height = asset.quality_width_height
        log_file_path = self._get_log_file_path(asset)

        yuv_type = self._get_workfile_yuv_type(asset)
        ref_path = asset.ref_procfile_path
        dis_path = asset.dis_procfile_path
        w = quality_width
        h = quality_height
        logger = self.logger

        optional_dict2 = self.optional_dict2 if self.optional_dict2 is not None else dict()

        ExternalProgramCaller.call_vmafexec_single_feature(
            'float_psnr', yuv_type, ref_path, dis_path, w, h,
            log_file_path, logger, options={**optional_dict2})


class MomentFeatureExtractor(FeatureExtractor):

    TYPE = "Moment_feature"

    # VERSION = "1.0" # call executable
    VERSION = "1.1"  # python only

    ATOM_FEATURES = ['ref1st', 'ref2nd', 'dis1st', 'dis2nd', ]

    DERIVED_ATOM_FEATURES = ['refvar', 'disvar', ]

    def _generate_result(self, asset):
        # routine to call the command-line executable and generate feature
        # scores in the log file.

        quality_w, quality_h = asset.quality_width_height

        with YuvReader(filepath=asset.ref_procfile_path, width=quality_w, height=quality_h,
                       yuv_type=self._get_workfile_yuv_type(asset)) as ref_yuv_reader:
            scores_mtx_list = []
            i = 0
            for ref_yuv in ref_yuv_reader:
                ref_y = ref_yuv[0]
                ref_y = ref_y.astype(np.double)
                firstm = ref_y.mean()
                secondm = ref_y.var() + firstm**2
                scores_mtx_list.append(np.hstack(([firstm], [secondm])))
                i += 1
            ref_scores_mtx = np.vstack(scores_mtx_list)

        with YuvReader(filepath=asset.dis_procfile_path, width=quality_w, height=quality_h,
                       yuv_type=self._get_workfile_yuv_type(asset)) as dis_yuv_reader:
            scores_mtx_list = []
            i = 0
            for dis_yuv in dis_yuv_reader:
                dis_y = dis_yuv[0]
                dis_y = dis_y.astype(np.double)
                firstm = dis_y.mean()
                secondm = dis_y.var() + firstm**2
                scores_mtx_list.append(np.hstack(([firstm], [secondm])))
                i += 1
            dis_scores_mtx = np.vstack(scores_mtx_list)

        assert ref_scores_mtx is not None and dis_scores_mtx is not None

        log_dict = {'ref_scores_mtx': ref_scores_mtx.tolist(),
                    'dis_scores_mtx': dis_scores_mtx.tolist()}

        log_file_path = self._get_log_file_path(asset)
        with open(log_file_path, 'wt') as log_file:
            log_file.write(str(log_dict))

    def _get_feature_scores(self, asset):
        # routine to read the feature scores from the log file, and return
        # the scores in a dictionary format.

        log_file_path = self._get_log_file_path(asset)

        with open(log_file_path, 'rt') as log_file:
            log_str = log_file.read()
            log_dict = ast.literal_eval(log_str)
        ref_scores_mtx = np.array(log_dict['ref_scores_mtx'])
        dis_scores_mtx = np.array(log_dict['dis_scores_mtx'])

        _, num_ref_features = ref_scores_mtx.shape
        assert num_ref_features == 2  # ref1st, ref2nd
        _, num_dis_features = dis_scores_mtx.shape
        assert num_dis_features == 2  # dis1st, dis2nd

        feature_result = {
            self.get_scores_key('ref1st'): list(ref_scores_mtx[:, 0]),
            self.get_scores_key('ref2nd'): list(ref_scores_mtx[:, 1]),
            self.get_scores_key('dis1st'): list(dis_scores_mtx[:, 0]),
            self.get_scores_key('dis2nd'): list(dis_scores_mtx[:, 1])
        }

        return feature_result

    @classmethod
    @override(Executor)
    def _post_process_result(cls, result):

        result = super(MomentFeatureExtractor, cls)._post_process_result(result)

        # calculate refvar and disvar from ref1st, ref2nd, dis1st, dis2nd
        refvar_scores_key = cls.get_scores_key('refvar')
        ref1st_scores_key = cls.get_scores_key('ref1st')
        ref2nd_scores_key = cls.get_scores_key('ref2nd')
        disvar_scores_key = cls.get_scores_key('disvar')
        dis1st_scores_key = cls.get_scores_key('dis1st')
        dis2nd_scores_key = cls.get_scores_key('dis2nd')
        get_var = lambda m: m[1] - m[0] * m[0]
        result.result_dict[refvar_scores_key] = \
            list(map(get_var, zip(result.result_dict[ref1st_scores_key],
                                  result.result_dict[ref2nd_scores_key])))
        result.result_dict[disvar_scores_key] = \
            list(map(get_var, zip(result.result_dict[dis1st_scores_key],
                                  result.result_dict[dis2nd_scores_key])))

        # validate
        for feature in cls.DERIVED_ATOM_FEATURES:
            assert cls.get_scores_key(feature) in result.result_dict

        return result


class SsimFeatureExtractor(VmafexecFeatureExtractorMixin, FeatureExtractor):

    TYPE = "SSIM_feature"
    # VERSION = "1.0"
    # VERSION = "1.1"  # fix OPT_RANGE_PIXEL_OFFSET = 0
    # VERSION = "1.2"  # call vmaf to replace standalone ssim exec
    VERSION = "1.3"  # add ssim_l, ssim_c, ssim_s as optional output

    ATOM_FEATURES = ['ssim',
                     'ssim_l', 'ssim_c', 'ssim_s',
                     ]

    ATOM_FEATURES_TO_VMAFEXEC_KEY_DICT = {
        'ssim': 'float_ssim',
        'ssim_l': 'float_ssim_l',
        'ssim_c': 'float_ssim_c',
        'ssim_s': 'float_ssim_s',
    }

    def _generate_result(self, asset):
        # routine to call the command-line executable and generate quality
        # scores in the log file.

        quality_width, quality_height = asset.quality_width_height
        log_file_path = self._get_log_file_path(asset)

        yuv_type = self._get_workfile_yuv_type(asset)
        ref_path = asset.ref_procfile_path
        dis_path = asset.dis_procfile_path
        w = quality_width
        h = quality_height
        logger = self.logger

        optional_dict2 = self.optional_dict2 if self.optional_dict2 is not None else dict()

        ExternalProgramCaller.call_vmafexec_single_feature(
            'float_ssim', yuv_type, ref_path, dis_path, w, h,
            log_file_path, logger, options={'enable_lcs': True, **optional_dict2})


class MsSsimFeatureExtractor(VmafexecFeatureExtractorMixin, FeatureExtractor):

    TYPE = "MS_SSIM_feature"
    # VERSION = "1.0"
    # VERSION = "1.1"  # fix OPT_RANGE_PIXEL_OFFSET = 0
    # VERSION = "1.2"  # call vmaf to replace standalone ms_ssim exec
    VERSION = "1.3"  # add ssim_l_scalex, ssim_c_scalex, ssim_s_scalex as optional output

    ATOM_FEATURES = ['ms_ssim',
                     'ms_ssim_l_scale0', 'ms_ssim_c_scale0', 'ms_ssim_s_scale0',
                     'ms_ssim_l_scale1', 'ms_ssim_c_scale1', 'ms_ssim_s_scale1',
                     'ms_ssim_l_scale2', 'ms_ssim_c_scale2', 'ms_ssim_s_scale2',
                     'ms_ssim_l_scale3', 'ms_ssim_c_scale3', 'ms_ssim_s_scale3',
                     'ms_ssim_l_scale4', 'ms_ssim_c_scale4', 'ms_ssim_s_scale4',
                     ]

    ATOM_FEATURES_TO_VMAFEXEC_KEY_DICT = {
        'ms_ssim': 'float_ms_ssim',

        'ms_ssim_l_scale0': 'float_ms_ssim_l_scale0',
        'ms_ssim_l_scale1': 'float_ms_ssim_l_scale1',
        'ms_ssim_l_scale2': 'float_ms_ssim_l_scale2',
        'ms_ssim_l_scale3': 'float_ms_ssim_l_scale3',
        'ms_ssim_l_scale4': 'float_ms_ssim_l_scale4',

        'ms_ssim_c_scale0': 'float_ms_ssim_c_scale0',
        'ms_ssim_c_scale1': 'float_ms_ssim_c_scale1',
        'ms_ssim_c_scale2': 'float_ms_ssim_c_scale2',
        'ms_ssim_c_scale3': 'float_ms_ssim_c_scale3',
        'ms_ssim_c_scale4': 'float_ms_ssim_c_scale4',

        'ms_ssim_s_scale0': 'float_ms_ssim_s_scale0',
        'ms_ssim_s_scale1': 'float_ms_ssim_s_scale1',
        'ms_ssim_s_scale2': 'float_ms_ssim_s_scale2',
        'ms_ssim_s_scale3': 'float_ms_ssim_s_scale3',
        'ms_ssim_s_scale4': 'float_ms_ssim_s_scale4',
    }

    def _generate_result(self, asset):
        # routine to call the command-line executable and generate quality
        # scores in the log file.

        quality_width, quality_height = asset.quality_width_height
        log_file_path = self._get_log_file_path(asset)

        yuv_type = self._get_workfile_yuv_type(asset)
        ref_path = asset.ref_procfile_path
        dis_path = asset.dis_procfile_path
        w = quality_width
        h = quality_height
        logger = self.logger

        optional_dict2 = self.optional_dict2 if self.optional_dict2 is not None else dict()

        ExternalProgramCaller.call_vmafexec_single_feature(
            'float_ms_ssim', yuv_type, ref_path, dis_path, w, h,
            log_file_path, logger, options={'enable_lcs': True, **optional_dict2})


class AnsnrFeatureExtractor(VmafexecFeatureExtractorMixin, FeatureExtractor):

    TYPE = "ANSNR_feature"
    VERSION = "1.0"

    ATOM_FEATURES = ['ansnr', 'anpsnr']

    ATOM_FEATURES_TO_VMAFEXEC_KEY_DICT = {
        'ansnr': 'float_ansnr',
        'anpsnr': 'float_anpsnr',
    }

    def _generate_result(self, asset):
        # routine to call the command-line executable and generate quality
        # scores in the log file.

        quality_width, quality_height = asset.quality_width_height
        log_file_path = self._get_log_file_path(asset)

        yuv_type = self._get_workfile_yuv_type(asset)
        ref_path = asset.ref_procfile_path
        dis_path = asset.dis_procfile_path
        w = quality_width
        h = quality_height
        logger = self.logger

        optional_dict2 = self.optional_dict2 if self.optional_dict2 is not None else dict()

        ExternalProgramCaller.call_vmafexec_single_feature(
            'float_ansnr', yuv_type, ref_path, dis_path, w, h, log_file_path, logger,
            options={**optional_dict2})



from __future__ import absolute_import

import json
import unittest
from functools import partial

import numpy as np

from vmaf.core.asset import Asset
from vmaf.config import VmafConfig
from vmaf.core.result import Result
from vmaf.core.result_store import FileSystemResultStore
from vmaf.core.quality_runner import VmafLegacyQualityRunner, VmafQualityRunner
from vmaf.tools.misc import MyTestCase
from vmaf.tools.stats import ListStats

from test.testutil import set_default_576_324_videos_for_testing

__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"


class ResultTest(MyTestCase):

    def setUp(self):
        super().setUp()
        ref_path = VmafConfig.test_resource_path("yuv", "checkerboard_1920_1080_10_3_0_0.yuv")
        dis_path = VmafConfig.test_resource_path("yuv", "checkerboard_1920_1080_10_3_1_0.yuv")
        asset = Asset(dataset="test", content_id=0, asset_id=0,
                      workdir_root=VmafConfig.workdir_path(),
                      ref_path=ref_path,
                      dis_path=dis_path,
                      asset_dict={'width': 1920, 'height': 1080})

        self.runner = VmafLegacyQualityRunner(
            [asset], None, fifo_mode=True,
            delete_workdir=True, result_store=FileSystemResultStore(),
        )
        self.runner.run()

        self.result = self.runner.results[0]

    def tearDown(self):
        if hasattr(self, 'runner'):
            self.runner.remove_results()
        super().tearDown()

    def test_todataframe_fromdataframe(self):

        df = self.result.to_dataframe()
        df_vmaf = df.loc[df['scores_key'] == 'VMAF_legacy_scores']
        df_adm = df.loc[df['scores_key'] == 'VMAF_feature_adm_scores']
        df_vif = df.loc[df['scores_key'] == 'VMAF_feature_vif_scores']
        df_ansnr = df.loc[df['scores_key'] == 'VMAF_feature_ansnr_scores']
        df_motion = df.loc[df['scores_key'] == 'VMAF_feature_motion_scores']
        df_adm_den = df.loc[df['scores_key'] == 'VMAF_feature_adm_den_scores']
        self.assertEqual(len(df), 38)
        self.assertEqual(len(df_vmaf), 1)
        self.assertEqual(len(df_adm), 1)
        self.assertEqual(len(df_vif), 1)
        self.assertEqual(len(df_ansnr), 1)
        self.assertEqual(len(df_motion), 1)
        self.assertAlmostEqual(np.mean(df_vmaf.iloc[0]['scores']), 40.421899030550769, places=4)
        self.assertAlmostEqual(np.mean(df_adm.iloc[0]['scores']), 0.78533833333333336, places=4)
        self.assertAlmostEqual(np.mean(df_vif.iloc[0]['scores']), 0.156834666667, places=4)
        self.assertAlmostEqual(np.mean(df_ansnr.iloc[0]['scores']), 7.92623066667, places=4)
        self.assertAlmostEqual(np.mean(df_motion.iloc[0]['scores']), 12.5548366667, places=4)
        self.assertAlmostEqual(np.mean(df_adm_den.iloc[0]['scores']), 2773.8912249999998, places=3)
        self.assertAlmostEqual(np.mean(Result.get_unique_from_dataframe(df, 'VMAF_legacy_scores', 'scores')), 40.421899030550769, places=4)
        self.assertAlmostEqual(np.mean(Result.get_unique_from_dataframe(df, 'VMAF_feature_adm_scores', 'scores')), 0.78533833333333336, places=4)
        self.assertAlmostEqual(np.mean(Result.get_unique_from_dataframe(df, 'VMAF_feature_vif_scores', 'scores')), 0.156834666667, places=4)
        self.assertAlmostEqual(np.mean(Result.get_unique_from_dataframe(df, 'VMAF_feature_ansnr_scores', 'scores')), 7.92623066667, places=4)
        self.assertAlmostEqual(np.mean(Result.get_unique_from_dataframe(df, 'VMAF_feature_motion_scores', 'scores')), 12.5548366667, places=4)
        self.assertEqual(df.iloc[0]['dataset'], 'test')
        self.assertEqual(df.iloc[0]['content_id'], 0)
        self.assertEqual(df.iloc[0]['asset_id'], 0)
        self.assertEqual(df.iloc[0]['ref_name'], 'checkerboard_1920_1080_10_3_0_0.yuv')
        self.assertEqual(df.iloc[0]['dis_name'], 'checkerboard_1920_1080_10_3_1_0.yuv')
        self.assertEqual(
            df.iloc[0]['asset'],
            '{"asset_dict": {"height": 1080, "use_path_as_workpath": 1, "use_workpath_as_procpath": 1, "width": 1920}, "asset_id": 0, "content_id": 0, "dataset": "test", "dis_path": "checkerboard_1920_1080_10_3_1_0.yuv", "ref_path": "checkerboard_1920_1080_10_3_0_0.yuv", "workdir": ""}') # noqa
        self.assertEqual(df.iloc[0]['executor_id'], 'VMAF_legacy_VF0.2.7-1.1')

        Result._assert_asset_dataframe(df)

        recon_result = Result.from_dataframe(df)
        self.assertEqual(self.result, recon_result)
        self.assertTrue(self.result == recon_result)
        self.assertFalse(self.result != recon_result)
        self.assertEqual(recon_result.asset.__class__, Asset)

    def test_to_score_str(self):
        self.assertAlmostEqual(self.result.get_result('VMAF_legacy_score'), 40.421899030550769, places=4)
        self.assertAlmostEqual(self.result['VMAF_legacy_score'], 40.421899030550769, places=4)
        self.assertAlmostEqual(self.result.get_result('VMAF_feature_adm_score'), 0.78533833333333336, places=4)
        self.assertAlmostEqual(self.result['VMAF_feature_adm_score'], 0.78533833333333336, places=4)
        self.assertAlmostEqual(self.result['VMAF_feature_vif_score'], 0.15683466666666665, places=4)
        self.assertAlmostEqual(self.result['VMAF_feature_motion_score'], 12.5548366667, places=4)
        self.assertAlmostEqual(self.result['VMAF_feature_ansnr_score'], 7.92623066667, places=4)
        self.result.set_score_aggregate_method(np.min)
        self.assertAlmostEqual(self.result.get_result('VMAF_legacy_score'), 37.573531379639725, places=4)
        self.result.set_score_aggregate_method(np.max)
        self.assertAlmostEqual(self.result.get_result('VMAF_legacy_score'), 44.815357234059327, places=4)
        self.result.set_score_aggregate_method(np.median)
        self.assertAlmostEqual(self.result.get_result('VMAF_legacy_score'), 38.876808477953254, places=4)
        self.result.set_score_aggregate_method(np.mean)
        self.assertAlmostEqual(self.result.get_result('VMAF_legacy_score'), 40.421899030550769, places=4)
        self.result.set_score_aggregate_method(np.std)
        self.assertAlmostEqual(self.result.get_result('VMAF_legacy_score'), 3.1518765879212993, places=4)
        self.result.set_score_aggregate_method(np.var)
        self.assertAlmostEqual(self.result.get_result('VMAF_legacy_score'), 9.9343260254864134, places=4)
        self.result.set_score_aggregate_method(partial(np.percentile, q=50))
        self.assertAlmostEqual(self.result.get_result('VMAF_legacy_score'), 38.876808477953254, places=4)
        self.result.set_score_aggregate_method(partial(np.percentile, q=80))
        self.assertAlmostEqual(self.result.get_result('VMAF_legacy_score'), 42.439937731616901, places=4)
        self.result.set_score_aggregate_method(ListStats.total_variation)
        self.assertAlmostEqual(self.result.get_result('VMAF_legacy_score'), 6.5901873052628375, places=4)
        self.result.set_score_aggregate_method(partial(ListStats.moving_average, n=2))
        for item in list(self.result.get_result('VMAF_legacy_score')):
            self.assertAlmostEqual(item, 42.86773029545747, places=6)

        with self.assertRaises(KeyError):
            self.result.get_result('VVMAF_legacy_score')
        with self.assertRaises(KeyError):
            self.result.get_result('VMAF_motion_score')


class ResultFormattingTest(unittest.TestCase):

    def setUp(self):

        # ref_path = VmafConfig.test_resource_path("yuv", "checkerboard_1920_1080_10_3_0_0.yuv")
        # dis_path = VmafConfig.test_resource_path("yuv", "checkerboard_1920_1080_10_3_1_0.yuv")
        # asset = Asset(dataset="test", content_id=0, asset_id=0,
        #               workdir_root=VmafConfig.workdir_path(),
        #               ref_path=ref_path,
        #               dis_path=dis_path,
        #               asset_dict={'width':1920, 'height':1080})
        #
        # self.runner = SsimQualityRunner(
        #     [asset], None, fifo_mode=True,
        #     delete_workdir=True, result_store=FileSystemResultStore(),
        # )
        # self.runner.run()
        #
        # FileSystemResultStore.save_result(self.runner.results[0], VmafConfig.test_resource_path('/ssim_result_for_test.txt')

        self.result = FileSystemResultStore.load_result(VmafConfig.test_resource_path('ssim_result_for_test.txt'))

    def tearDown(self):
        # if hasattr(self, 'runner'):
        #     self.runner.remove_results()
        pass

    @unittest.skip("numerical value has changed.")
    def test_to_xml(self):
        self.assertEqual(self.result.to_xml().strip(), """
<?xml version="1.0" ?>
<result executorId="SSIM_V1.0">
  <asset identifier="test_0_0_checkerboard_1920_1080_10_3_0_0_1920x1080_vs_checkerboard_1920_1080_10_3_1_0_1920x1080_q_1920x1080"/>
  <frames>
    <frame SSIM_feature_ssim_c_score="0.997404" SSIM_feature_ssim_l_score="0.999983" SSIM_feature_ssim_s_score="0.935802" SSIM_score="0.933353" frameNum="0"/>
    <frame SSIM_feature_ssim_c_score="0.997404" SSIM_feature_ssim_l_score="0.999983" SSIM_feature_ssim_s_score="0.935802" SSIM_score="0.933353" frameNum="1"/>
    <frame SSIM_feature_ssim_c_score="0.997404" SSIM_feature_ssim_l_score="0.999983" SSIM_feature_ssim_s_score="0.935803" SSIM_score="0.933354" frameNum="2"/>
  </frames>
  <aggregate SSIM_feature_ssim_c_score="0.997404" SSIM_feature_ssim_l_score="0.999983" SSIM_feature_ssim_s_score="0.935802333333" SSIM_score="0.933353333333" method="mean"/>
</result>
        """.strip())

    def test_to_json(self):
        self.assertEqual(self.result.to_dict(), json.loads("""
{
    "executorId": "SSIM_V1.0",
    "asset": {
        "identifier": "test_0_0_checkerboard_1920_1080_10_3_0_0_1920x1080_vs_checkerboard_1920_1080_10_3_1_0_1920x1080_q_1920x1080"
    },
    "frames": [
        {
            "frameNum": 0,
            "SSIM_feature_ssim_c_score": 0.997404,
            "SSIM_feature_ssim_l_score": 0.999983,
            "SSIM_feature_ssim_s_score": 0.935802,
            "SSIM_score": 0.933353
        },
        {
            "frameNum": 1,
            "SSIM_feature_ssim_c_score": 0.997404,
            "SSIM_feature_ssim_l_score": 0.999983,
            "SSIM_feature_ssim_s_score": 0.935802,
            "SSIM_score": 0.933353
        },
        {
            "frameNum": 2,
            "SSIM_feature_ssim_c_score": 0.997404,
            "SSIM_feature_ssim_l_score": 0.999983,
            "SSIM_feature_ssim_s_score": 0.935803,
            "SSIM_score": 0.933354
        }
    ],
    "aggregate": {
        "SSIM_feature_ssim_c_score": 0.99740399999999996,
        "SSIM_feature_ssim_l_score": 0.99998299999999996,
        "SSIM_feature_ssim_s_score": 0.93580233333333329,
        "SSIM_score": 0.93335333333333337,
        "method": "mean"
    }
}
        """))


class ResultStoreTest(unittest.TestCase):

    def setUp(self):
        ref_path = VmafConfig.test_resource_path("yuv", "checkerboard_1920_1080_10_3_0_0.yuv")
        dis_path = VmafConfig.test_resource_path("yuv", "checkerboard_1920_1080_10_3_1_0.yuv")
        asset = Asset(dataset="test", content_id=0, asset_id=0,
                      workdir_root=VmafConfig.workdir_path(),
                      ref_path=ref_path,
                      dis_path=dis_path,
                      asset_dict={'width':1920, 'height':1080})

        self.runner = VmafLegacyQualityRunner(
            [asset], None, fifo_mode=True,
            delete_workdir=True, result_store=None,
        )
        self.runner.run()
        self.result = self.runner.results[0]

    def tearDown(self):
        if hasattr(self, 'result') and hasattr(self, 'result_store'):
            self.result_store.delete(self.result.asset, self.result.executor_id)
        pass

    def test_file_system_result_store_save_load(self):
        self.result_store = FileSystemResultStore(logger=None)
        asset = self.result.asset
        executor_id = self.result.executor_id

        self.result_store.save(self.result)

        loaded_result = self.result_store.load(asset, executor_id)

        self.assertEqual(self.result, loaded_result)


class ResultStoreTestWithNone(unittest.TestCase):

    def test_load_result_with_none(self):

        result = FileSystemResultStore.load_result(VmafConfig.test_resource_path('result_with_none.txt'))
        result.set_score_aggregate_method(ListStats.nonemean)
        self.assertAlmostEqual(result['STRRED_feature_srred_score'], 5829.2644469999996, places=4)
        self.assertEqual(result.asset.__class__, Asset)


class ResultAggregatingTest(unittest.TestCase):

    def test_from_xml_from_json_and_aggregation(self):

        ref_path, dis_path, asset, asset_original = set_default_576_324_videos_for_testing()

        asset_list = [asset, asset_original]

        self.runner = VmafQualityRunner(
            asset_list,
            None, fifo_mode=True,
            delete_workdir=True,
            result_store=None,
            optional_dict={
                'model_filepath': VmafConfig.model_path("vmaf_float_v0.6.1.pkl"),
            },
            optional_dict2=None,
        )
        self.runner.run()

        results = self.runner.results

        xml_string_expected = results[0].to_xml()
        xml_string_recon = Result.from_xml(xml_string_expected).to_xml()

        json_string_expected = results[0].to_json()
        json_string_recon = Result.from_json(json_string_expected).to_json()

        assert xml_string_expected == xml_string_recon, "XML files do not match"
        assert json_string_expected == json_string_recon, "JSON files do not match"

        combined_result = Result.combine_result([results[0], results[1]])

        # check that all keys are there
        combined_result_keys = [key for key in combined_result.result_dict]
        keys_0 = [key for key in results[0].result_dict]
        keys_1 = [key for key in results[1].result_dict]
        assert set(keys_0) == set(keys_1) == set(combined_result_keys)

        # check that the dictionaries have been copied as expected
        for key in combined_result_keys:
            assert len(combined_result.result_dict[key]) == len(results[0].result_dict[key]) + len(results[1].result_dict[key])
            assert combined_result.result_dict[key][0] == results[0].result_dict[key][0]
            assert combined_result.result_dict[key][len(results[0].result_dict[key]) - 1] == results[0].result_dict[key][len(results[0].result_dict[key]) - 1]
            assert combined_result.result_dict[key][len(results[0].result_dict[key])] == results[1].result_dict[key][0]
            assert combined_result.result_dict[key][len(combined_result.result_dict[key]) - 1] == results[1].result_dict[key][len(results[1].result_dict[key]) - 1]


class ScoreAggregationTest(unittest.TestCase):

    def setUp(self):

        ref_path = VmafConfig.test_resource_path("yuv", "checkerboard_1920_1080_10_3_0_0.yuv")
        dis_path = VmafConfig.test_resource_path("yuv", "checkerboard_1920_1080_10_3_1_0.yuv")
        asset = Asset(dataset="test", content_id=0, asset_id=0,
                      workdir_root=VmafConfig.workdir_path(),
                      ref_path=ref_path,
                      dis_path=dis_path,
                      asset_dict={'width':1920, 'height':1080})

        self.runner = VmafQualityRunner(
            [asset], None, fifo_mode=True,
            delete_workdir=True, result_store=FileSystemResultStore(),
            optional_dict={
                'model_filepath': VmafConfig.model_path("vmaf_float_v0.6.1.json"),
            },
        )
        self.runner.run()

        self.result = self.runner.results[0]

        Nframes = len(self.result.result_dict['VMAF_scores'])
        self.result.result_dict['VMAF_array_scores'] = self.result.result_dict['VMAF_scores'].reshape(Nframes, 1)
        self.result.result_dict['VMAF_two_models_array_scores'] = np.vstack((self.result.result_dict['VMAF_scores'].reshape(1, Nframes),
                                                                       self.result.result_dict['VMAF_scores'].reshape(1, Nframes)))
        self.result.result_dict['VMAF_3D_array_scores'] = np.zeros((1, 1, 1))

    def tearDown(self):
        if hasattr(self, 'runner'):
            self.runner.remove_results()

    def test_to_score_str(self):

        self.result.set_score_aggregate_method(np.mean)
        # the following should give same value
        self.assertAlmostEqual(self.result['VMAF_score'], 35.0661575902223, places=4)
        # for a 2-D array, first dimension is # models and second is # frames
        self.assertAlmostEqual(self.result['VMAF_two_models_array_score'][0], 35.0661575902223, places=4)
        self.assertAlmostEqual(self.result['VMAF_two_models_array_score'][1], 35.0661575902223, places=4)
        self.assertAlmostEqual(self.result['VMAF_array_score'][0], 22.97749190550349, places=4)
        self.assertAlmostEqual(self.result['VMAF_array_score'][1], 44.79653061901706, places=4)
        self.assertAlmostEqual(self.result['VMAF_array_score'][2], 37.424450246146364, places=4)
        # check that a 3-D array will throw assertion, score aggregation accepts only up to 2-D
        with self.assertRaises(AssertionError):
            x = self.result['VMAF_3D_array_score']


if __name__ == '__main__':
    unittest.main(verbosity=2)



dataset_name = 'example_image'

yuv_fmt = 'notyuv'
quality_width = 1920
quality_height = 1080
workfile_yuv_type = 'yuv444p'

from vmaf.config import VmafConfig

ref_videos = [
    {'content_id': 0, 'path': VmafConfig.resource_path('icpf', 'frame00000001.icpf')},

    {'content_id': 1, 'path': VmafConfig.resource_path('icpf', 'frame00000002.icpf')},
]

dis_videos = [
    {'content_id': 0, 'asset_id': 0, 'dmos': 100, 'path': VmafConfig.resource_path('icpf', 'frame00000000.icpf')}, # ref
    {'content_id': 0, 'asset_id': 1, 'dmos': 50,  'path': VmafConfig.resource_path('icpf', 'frame00000001.icpf')},

    {'content_id': 1, 'asset_id': 2, 'dmos': 100,  'path': VmafConfig.resource_path('icpf', 'frame00000002.icpf')}, # ref
    {'content_id': 1, 'asset_id': 3, 'dmos': 80,  'path': VmafConfig.resource_path('icpf', 'frame00000003.icpf')},
]





#!/usr/bin/env python3

import argparse
import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), '..', '..'))

from vmaf.core.train_test_model import LibsvmNusvrTrainTestModel


if __name__ == '__main__':

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--input-pkl-filepath", dest="input_pkl_filepath", nargs=1, type=str,
        help="path to the input pkl file, example: model/vmaf_float_v0.6.1.pkl or model/vmaf_float_b_v0.6.3/vmaf_float_b_v0.6.3.pkl", required=True)

    parser.add_argument(
        "--output-json-filepath", dest="output_json_filepath", nargs=1, type=str,
        help="path to the output json file, example: model/vmaf_float_v0.6.1.json or model/vmaf_float_b_v0.6.3.json", required=True)

    args = parser.parse_args()

    input_pkl_filepath = args.input_pkl_filepath[0]
    output_json_filepath = args.output_json_filepath[0]

    loaded_model = LibsvmNusvrTrainTestModel.from_file(input_pkl_filepath, None, format='pkl')
    loaded_model.to_file(output_json_filepath, format='json', combined=True)

    print(f'converted {loaded_model.TYPE} model from pkl file {input_pkl_filepath} to json file {output_json_filepath}')

    exit(0)



import numpy as np

__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"


class InterpolationUtils(object):

    # /**
    #  * Compute PCHIP params and compute the rate from distortion (or PSNR).
    #  *
    #  * @param rdPointsList        List of RD points, minimum 4 needed
    #  * @param arrayDistInLogScale List of distortion values for which rate values are desired
    #  * @return List of rate values, either in log scale or linear scale
    #  */
    @classmethod
    def interpolateRateFromMetric(cls, rdPointsList, arrayDistInLogScale):

        N = len(rdPointsList)

        log_rate = []
        log_dist = []
        H = []
        delta = []
        d = []
        c = []
        b = []

        cls.computeParamsForSegments(rdPointsList, log_rate, log_dist, H, delta, d, c, b, False)

        arrayRate = []

        for distInLogScale in arrayDistInLogScale:
            distBegin = log_dist[0]
            yk = log_rate[0]
            segmentOfInterest = 0
            for i in range(N-1):
                if log_dist[i] <= distInLogScale <= log_dist[i + 1]:
                    distBegin = log_dist[i]
                    yk = log_rate[i]
                    segmentOfInterest = i
                    break

            rate = cls.computeRate(distInLogScale, distBegin, yk, d[segmentOfInterest], c[segmentOfInterest], b[segmentOfInterest])
            rate = min(rate, log_rate[segmentOfInterest + 1])
            rate = max(rate, log_rate[segmentOfInterest]) # make sure interpolated rate is within bounds
            arrayRate.append(rate)

        return arrayRate

    # /***
    #  *
    #  * Compute rate from distortion using formula. Use parameters for the curve segment that are passed in.
    #  *
    #  * @param x     distortion for which rate needs to be computed
    #  * @param xk    distortion at the beginning of curve segment
    #  * @param yk    rate (in log scale) at beginning of curve segment
    #  * @param dk    d parameter of relevant curve segment
    #  * @param ck    c parameter of relevant curve segment
    #  * @param bk    b parameter of relevant curve segment
    #  * @return rate in log scale computed using formula and passed in parameters
    #  */
    @staticmethod
    def computeRate(x, xk, yk, dk, ck, bk):

        s = x - xk

        return yk + (s * dk) + (s * s * ck) + (s * s * s * bk)

    @classmethod
    def computeParamsForSegments(cls, rdPointsList, log_rate, log_dist, H, delta, d, c, b, convertRateUsingLogarithm):
        N = len(rdPointsList)

        for i in range(N):
            if convertRateUsingLogarithm:
                log_rate.append(np.log10(rdPointsList[i][0]))
            else:
                log_rate.append(rdPointsList[i][0])
            log_dist.append(rdPointsList[i][1])

        # // H_i = length of i_th subinterval
        # // delta_i = slope of i_th subinterval
        for i in range(N - 1):
            H.append(log_dist[i + 1] - log_dist[i])
            delta.append((log_rate[i + 1] - log_rate[i]) / H[i])

        # // Determine slopes at each point = coefficient of first-order term

        # // determine slope at starting point
        d.append(cls.pchipend(H[0], H[1], delta[0], delta[1]))

        # // determine slope at all intermediate points: slope is weighted harmonic mean of two slopes (delta_{i-1} and delta_i)
        for i in range(1, N-1):
            d.append((3 * H[i - 1] + 3 * H[i]) / ((2 * H[i] + H[i - 1]) / delta[i - 1] + (H[i] + 2 * H[i - 1]) / delta[i]))

        # // determine slope at end point
        d.append(cls.pchipend(H[N - 2], H[N - 3], delta[N - 2], delta[N - 3]))

        # // determine second- and third-order coefficients for each segment

        for i in range(N - 1):
            c.append((3 * delta[i] - 2 * d[i] - d[i + 1]) / H[i])
            b.append((d[i] - 2 * delta[i] + d[i + 1]) / (H[i] * H[i]))

    @staticmethod
    def pchipend(h1, h2, delta1, delta2):
        # // one-sided formula for endpoints
        # // non-centered shape-preserving three point formula
        d = ((2 * h1 + h2) * delta1 - h1 * delta2) / (h1 + h2)

        # // opposite slopes -> set slope to zero
        if d * delta1 < 0:
            d = 0
        elif (delta1 * delta2 < 0) and (abs(d) > abs(3 * delta1)):
            d = 3 * delta1

        return d



import os
import shutil
import unittest
from test.testutil import set_default_576_324_videos_for_testing, \
    set_default_576_324_videos_for_testing_scaled, \
    set_default_cambi_video_for_testing_b, \
    set_default_cambi_video_for_testing_10b

from vmaf.core.cambi_feature_extractor import CambiFeatureExtractor, CambiFullReferenceFeatureExtractor
from vmaf.core.cambi_quality_runner import CambiQualityRunner, CambiFullReferenceQualityRunner
from vmaf.tools.misc import MyTestCase
from vmaf.core.result_store import FileSystemResultStore


class CambiFeatureExtractorTest(MyTestCase):

    def tearDown(self):
        if hasattr(self, 'fextractor'):
            self.fextractor.remove_results()
        super().tearDown()

    def test_run_cambi_fextractor(self):
        _, _, asset, asset_original = set_default_576_324_videos_for_testing()
        self.fextractor = CambiFeatureExtractor(
            [asset, asset_original],
            None, fifo_mode=False,
            result_store=None
        )
        self.fextractor.run(parallelize=True)
        results = self.fextractor.results

        # score: arithmetic mean score over all frames
        self.assertAlmostEqual(results[0]['Cambi_feature_cambi_score'],
                               0.25968416666666666, places=4)
        self.assertAlmostEqual(results[1]['Cambi_feature_cambi_score'],
                               0.00020847916666666663, places=4)

    def test_run_cambi_fextractor_scaled(self):
        _, _, asset, asset_original = set_default_576_324_videos_for_testing_scaled()
        self.fextractor = CambiFeatureExtractor(
            [asset, asset_original],
            None, fifo_mode=False,
            result_store=None,
            optional_dict={}
        )
        self.fextractor.run(parallelize=True)
        results = self.fextractor.results

        # score: arithmetic mean score over all frames
        self.assertAlmostEqual(results[0]['Cambi_feature_cambi_score'],
                               0.17871631249999997, places=4)
        self.assertAlmostEqual(results[1]['Cambi_feature_cambi_score'],
                               0.00022027083333333336, places=4)

    def test_run_cambi_fextractor_scaled_b(self):
        _, _, asset, asset_original = set_default_cambi_video_for_testing_b()
        self.fextractor = CambiFeatureExtractor(
            [asset, asset_original],
            None, fifo_mode=False,
            result_store=None,
            optional_dict={}
        )
        self.fextractor.run(parallelize=True)
        results = self.fextractor.results

        # score: arithmetic mean score over all frames
        self.assertAlmostEqual(results[0]['Cambi_feature_cambi_score'],
                               0.773078, places=4)

    def test_run_cambi_fextractor_10b(self):
        _, _, asset, asset_original = set_default_cambi_video_for_testing_10b()
        self.fextractor = CambiFeatureExtractor(
            [asset, asset_original],
            None, fifo_mode=False,
            result_store=None,
            optional_dict={}
        )
        self.fextractor.run(parallelize=True)
        results = self.fextractor.results

        # score: arithmetic mean score over all frames
        self.assertAlmostEqual(results[0]['Cambi_feature_cambi_score'],
                               0.0013863333333333334, places=4)

    def test_run_cambi_fextractor_incorrect_enc_bitdepth(self):
        _, _, asset, asset_original = set_default_cambi_video_for_testing_10b()
        asset.asset_dict['dis_enc_bitdepth'] = 10
        self.fextractor = CambiFeatureExtractor(
            [asset, asset_original],
            None, fifo_mode=False,
            result_store=None,
            optional_dict={}
        )
        self.fextractor.run(parallelize=False)
        results = self.fextractor.results

        # score: arithmetic mean score over all frames
        self.assertAlmostEqual(results[0]['Cambi_feature_cambi_score'],
                               0.0013863333333333334, places=4)

    def test_run_cambi_fextractor_correct_enc_bitdepth(self):
        _, _, asset, asset_original = set_default_cambi_video_for_testing_10b()
        asset.asset_dict['dis_enc_bitdepth'] = 8
        self.fextractor = CambiFeatureExtractor(
            [asset, asset_original],
            None, fifo_mode=False,
            result_store=None,
            optional_dict={}
        )
        self.fextractor.run(parallelize=False)
        results = self.fextractor.results

        # score: arithmetic mean score over all frames
        self.assertAlmostEqual(results[0]['Cambi_feature_cambi_score'],
                               0.00020733333333333332, places=4)

    def test_run_cambi_fextractor_notyuv_correct_enc_bitdepth_8(self):
        _, _, asset, asset_original = set_default_576_324_videos_for_testing()
        self.fextractor = CambiFeatureExtractor(
            [asset, asset_original],
            None, fifo_mode=False,
            result_store=None,
            optional_dict={}
        )
        self.fextractor.run(parallelize=False)
        results = self.fextractor.results

        # score: arithmetic mean score over all frames
        self.assertAlmostEqual(results[1]['Cambi_feature_cambi_score'],
                               0.00020733333333333332, places=4)

    def test_run_cambi_fextractor_enc_bitdepth_none(self):
        _, _, asset, asset_original = set_default_cambi_video_for_testing_10b()
        asset.asset_dict['dis_enc_bitdepth'] = None
        self.fextractor = CambiFeatureExtractor(
            [asset, asset_original],
            None, fifo_mode=False,
            result_store=None,
            optional_dict={}
        )
        with self.assertRaises(AssertionError):
            self.fextractor.run(parallelize=False)

    def test_run_cambi_fextractor_max_log_contrast(self):
        _, _, asset, asset_original = set_default_576_324_videos_for_testing()
        self.fextractor = CambiFeatureExtractor(
            [asset, asset_original],
            None, fifo_mode=False,
            result_store=None,
            optional_dict={'max_log_contrast': 4}
        )
        self.fextractor.run(parallelize=True)
        results = self.fextractor.results

        # score: arithmetic mean score over all frames
        self.assertAlmostEqual(results[0]['Cambi_feature_cambi_score'],
                               0.32247770833333333, places=4)
        self.assertAlmostEqual(results[1]['Cambi_feature_cambi_score'],
                               0.0002582708333333333, places=4)

        self.fextractor = CambiFeatureExtractor(
            [asset, asset_original],
            None, fifo_mode=False,
            result_store=None,
            optional_dict={'max_log_contrast': 0}
        )
        self.fextractor.run(parallelize=True)
        results = self.fextractor.results

        # score: arithmetic mean score over all frames
        self.assertAlmostEqual(results[0]['Cambi_feature_cambi_score'],
                               0.0024372291666666665, places=4)
        self.assertAlmostEqual(results[1]['Cambi_feature_cambi_score'],
                               0.00019045833333333336, places=4)

    def test_run_cambi_fextractor_full_reference(self):
        _, _, asset, asset_original = set_default_576_324_videos_for_testing()
        self.fextractor = CambiFullReferenceFeatureExtractor(
            [asset, asset_original],
            None, fifo_mode=False,
            result_store=None,
        )
        self.fextractor.run(parallelize=True)
        results = self.fextractor.results

        # score: arithmetic mean score over all frames
        self.assertAlmostEqual(results[0]['Cambi_FR_feature_cambi_score'],
                               0.25968416666666666, places=4)
        self.assertAlmostEqual(results[0]['Cambi_FR_feature_cambi_source_score'],
                               0.00020847916666666663, places=4)
        self.assertAlmostEqual(results[0]['Cambi_FR_feature_cambi_full_reference_score'],
                               0.2594353333333333, places=4)

    def test_run_cambi_fextractor_full_reference_scaled_ref(self):
        _, _, asset, asset_original = set_default_576_324_videos_for_testing()
        self.fextractor = CambiFullReferenceFeatureExtractor(
            [asset, asset_original],
            None, fifo_mode=False,
            result_store=None,
            optional_dict={'src_width': 480, 'src_height': 270}
        )
        self.fextractor.run(parallelize=True)
        results = self.fextractor.results

        # score: arithmetic mean score over all frames
        self.assertAlmostEqual(results[0]['Cambi_FR_feature_cambi_score'],
                               0.25968416666666666, places=4)
        self.assertAlmostEqual(results[0]['Cambi_FR_feature_cambi_source_score'],
                               0.00022027083333333336, places=4)
        self.assertAlmostEqual(results[0]['Cambi_FR_feature_cambi_full_reference_score'],
                               0.25944045833333335, places=4)


class CambiQualityRunnerTest(MyTestCase):

    def test_run_cambi_runner(self):
        _, _, asset, asset_original = set_default_576_324_videos_for_testing()
        self.qrunner = CambiQualityRunner(
            [asset, asset_original],
            None, fifo_mode=False,
            result_store=None
        )
        self.qrunner.run(parallelize=True)
        results = self.qrunner.results

        # score: arithmetic mean score over all frames
        self.assertAlmostEqual(results[0]['Cambi_score'],
                               0.25968416666666666, places=4)
        self.assertAlmostEqual(results[1]['Cambi_score'],
                               0.00020847916666666663, places=4)

    def test_run_cambi_runner_scale(self):
        _, _, asset, asset_original = set_default_576_324_videos_for_testing_scaled()
        self.qrunner = CambiQualityRunner(
            [asset, asset_original],
            None, fifo_mode=False,
            result_store=None,
            optional_dict={}
        )
        self.qrunner.run(parallelize=True)
        results = self.qrunner.results

        # score: arithmetic mean score over all frames
        self.assertAlmostEqual(results[0]['Cambi_score'],
                               0.17871631249999997, places=4)
        self.assertAlmostEqual(results[1]['Cambi_score'],
                               0.00022027083333333336, places=4)

    def test_run_cambi_runner_scale_b(self):
        _, _, asset, asset_original = set_default_cambi_video_for_testing_b()
        self.qrunner = CambiQualityRunner(
            [asset, asset_original],
            None, fifo_mode=False,
            result_store=None,
            optional_dict={}
        )
        self.qrunner.run(parallelize=True)
        results = self.qrunner.results

        # score: arithmetic mean score over all frames
        self.assertAlmostEqual(results[0]['Cambi_score'],
                               0.773078, places=4)

    def test_run_cambi_runner_10b(self):
        _, _, asset, asset_original = set_default_cambi_video_for_testing_10b()
        self.qrunner = CambiQualityRunner(
            [asset, asset_original],
            None, fifo_mode=False,
            result_store=None,
            optional_dict={}
        )
        self.qrunner.run(parallelize=True)
        results = self.qrunner.results

        # score: arithmetic mean score over all frames
        self.assertAlmostEqual(results[0]['Cambi_score'],
                               0.0013863333333333334, places=4)

    def test_run_cambi_runner_fullref(self):
        _, _, asset, asset_original = set_default_576_324_videos_for_testing()
        self.qrunner = CambiFullReferenceQualityRunner(
            [asset, asset_original],
            None, fifo_mode=False,
            result_store=None,
        )
        self.qrunner.run(parallelize=True)
        results = self.qrunner.results

        # score: arithmetic mean score over all frames
        self.assertAlmostEqual(results[0]['Cambi_FR_score'],
                               0.2594353333333333, places=4)
        self.assertAlmostEqual(results[0]['Cambi_FR_feature_cambi_score'],
                               0.25968416666666666, places=4)


class CambiResultsCachingTest(MyTestCase):

    def setUp(self):
        self.results_store_dir = FileSystemResultStore()

    def tearDown(self):
        if os.path.exists(self.store_dir):
            shutil.rmtree(self.store_dir)

    def test_run_cambi_runner(self):
        _, _, asset, asset_original = set_default_576_324_videos_for_testing_scaled()
        self.qrunner = CambiQualityRunner(
            [asset, asset_original],
            None, fifo_mode=False,
            result_store=self.results_store_dir,
            optional_dict={}
        )

        # make sure the caching directory needs to be created by running the feature extractor
        fextractor= self.qrunner._get_feature_extractor_class()
        self.store_dir = os.path.join(
            self.results_store_dir.result_store_dir, f'{fextractor.TYPE}_V{fextractor.VERSION}', 'test')
        if os.path.exists(self.store_dir):
            shutil.rmtree(self.store_dir)

        self.qrunner.run(parallelize=False)
        results = self.qrunner.results

        # check if the correct directory was created
        self.assertTrue(os.path.exists(self.store_dir))

        # score: arithmetic mean score over all frames
        self.assertAlmostEqual(results[0]['Cambi_score'],
                               0.17871631249999997, places=4)
        self.assertAlmostEqual(results[1]['Cambi_score'],
                               0.00022027083333333336, places=4)


if __name__ == '__main__':
    unittest.main(verbosity=2)


from vmaf.tools.misc import unroll_dict_of_lists

__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"

import random
from math import floor
import sys


class ModelCrossValidation(object):

    @staticmethod
    def run_cross_validation(train_test_model_class,
                             model_param,
                             results_or_df,
                             train_indices,
                             test_indices,
                             optional_dict2=None
                             ):
        """
        Simple cross validation.
        :param train_test_model_class:
        :param model_param:
        :param results_or_df: list of BasicResult, or pandas.DataFrame
        :param train_indices:
        :param test_indices:
        :return:
        """
        xys_train = train_test_model_class.get_xys_from_results(results_or_df, train_indices)
        xs_test = train_test_model_class.get_xs_from_results(results_or_df, test_indices)
        ys_test = train_test_model_class.get_ys_from_results(results_or_df, test_indices)

        model = train_test_model_class(model_param,
                                       logger=None,
                                       optional_dict2=optional_dict2)
        model.train(xys_train)
        stats = model.evaluate(xs_test, ys_test)

        output = {'stats': stats, 'model': model,
                  'contentids': ys_test['content_id']}

        return output

    @classmethod
    def run_kfold_cross_validation(cls,
                                   train_test_model_class,
                                   model_param,
                                   results_or_df,
                                   kfold,
                                   logger=None,
                                   optional_dict2=None):
        """
        Standard k-fold cross validation, given hyperparameter set model_param
        :param train_test_model_class:
        :param model_param:
        :param results_or_df: list of BasicResult, or pandas.DataFrame
        :param kfold: if it is an integer, it is the number of folds; if it is
        list of indices, then each list contains row indices of the dataframe
        selected as one fold
        :return: output
        """

        if isinstance(kfold, int):
            kfold_type = 'int'
        elif isinstance(kfold, (list, tuple)):
            kfold_type = 'list'
        else:
            assert False, 'kfold must be either a list of lists or an integer.'

        # if input is integer (e.g. 4), reconstruct kfold in list of indices
        # format
        if kfold_type == 'int':
            num_fold = kfold
            dataframe_size = len(results_or_df)
            fold_size = int(floor(dataframe_size / num_fold))
            kfold = []
            for fold in range(num_fold):
                index_start = fold * fold_size
                index_end = min((fold+1)*fold_size, dataframe_size)
                kfold.append(range(index_start, index_end))

        assert len(kfold) >= 2, 'kfold list must have length >= 2 for k-fold ' \
                                'cross validation.'

        statss = []
        models = []
        contentids = []

        for fold in range(len(kfold)):

            # to avoid interference among folds
            if hasattr(train_test_model_class, 'reset'):
                train_test_model_class.reset()

            if logger:
                logger.info("Fold {}...".format(fold))

            test_index_range = kfold[fold]
            train_index_range = []
            for train_fold in range(len(kfold)):
                if train_fold != fold:
                    train_index_range += kfold[train_fold]

            output = cls.run_cross_validation(train_test_model_class,
                                              model_param,
                                              results_or_df,
                                              train_index_range,
                                              test_index_range,
                                              optional_dict2)

            stats = output['stats']
            model = output['model']

            statss.append(stats)
            models.append(model)

            contentids += list(output['contentids'])

        aggr_stats = train_test_model_class.aggregate_stats_list(statss)

        output = {'aggr_stats': aggr_stats, 'statss': statss, 'models': models}

        assert contentids is not None
        output['contentids'] = contentids

        return output

    @classmethod
    def run_nested_kfold_cross_validation(cls,
                                          train_test_model_class,
                                          model_param_search_range,
                                          results_or_df,
                                          kfold,
                                          search_strategy='grid',
                                          random_search_times=100,
                                          logger=None,
                                          optional_dict2=None,
                                          ):
        """
        Nested k-fold cross validation, given hyperparameter search range. The
        search range is specified in the format of, e.g.:
        {'norm_type':['normalize', 'clip_0to1', 'clip_minus1to1'],
         'n_estimators':[10, 50],
         'random_state': [0]}
        :param train_test_model_class:
        :param model_param_search_range:
        :param results_or_df: list of BasicResult, or pandas.DataFrame
        :param kfold: if it is an integer, it is the number of folds; if it is
        lists of indices, then each list contains row indices of the dataframe
        selected as one fold
        :param search_strategy: either 'grid' or 'random'
        :return: output
        """

        if isinstance(kfold, int):
            kfold_type = 'int'
        elif isinstance(kfold, (list, tuple)):
            kfold_type = 'list'
        else:
            assert False, 'kfold must be either a list of lists or an integer.'

        # if input is integer (e.g. 4), reconstruct kfold in list of indices
        # format
        if kfold_type == 'int':
            num_fold = kfold
            dataframe_size = len(results_or_df)
            fold_size = int(floor(dataframe_size / num_fold))
            kfold = []
            for fold in range(num_fold):
                index_start = fold * fold_size
                index_end = min((fold+1)*fold_size, dataframe_size)
                kfold.append(range(index_start, index_end))

        assert len(kfold) >= 3, 'kfold list must have length >= 2 for nested ' \
                                'k-fold cross validation.'

        if search_strategy == 'grid':
            cls._assert_grid_search(model_param_search_range)
            list_model_param = unroll_dict_of_lists(
                model_param_search_range)
        elif search_strategy == 'random':
            cls._assert_random_search(model_param_search_range)
            list_model_param = cls._sample_model_param_list(
                model_param_search_range, random_search_times)
        else:
            assert False, "Unknown search_strategy: {}".format(search_strategy)

        statss = []
        model_params = []
        contentids = []

        for fold in range(len(kfold)):

            if logger:
                logger.info("Fold {}...".format(fold))

            test_index_range = kfold[fold]
            train_index_range = []
            train_index_range_in_list_of_indices = []

            # in this case, train_index_range is list of lists
            for train_fold in range(len(kfold)):
                if train_fold != fold:
                    train_index_range += kfold[train_fold]
                    train_index_range_in_list_of_indices.append(kfold[train_fold])

            # iterate through all possible combinations of model_params
            best_model_param = None
            best_stats = None
            for model_param in list_model_param:

                if logger:
                    logger.info("\tModel parameter: {}".format(model_param))

                output = \
                    cls.run_kfold_cross_validation(train_test_model_class,
                                                   model_param,
                                                   results_or_df,
                                                   train_index_range_in_list_of_indices,
                                                   optional_dict2)
                stats = output['aggr_stats']

                if (best_stats is None) or (
                    train_test_model_class.get_objective_score(stats, score_type='SRCC')
                    >
                    train_test_model_class.get_objective_score(best_stats, score_type='SRCC')
                ):
                    best_stats = stats
                    best_model_param = model_param

            # run cross validation based on best model parameters
            output_ = cls.run_cross_validation(
                train_test_model_class, best_model_param, results_or_df,
                train_index_range, test_index_range, optional_dict2
            )
            stats_ = output_['stats']

            statss.append(stats_)
            model_params.append(best_model_param)

            contentids += list(output_['contentids'])

        aggr_stats = train_test_model_class.aggregate_stats_list(statss)
        top_model_param, count = cls._find_most_frequent_dict(model_params)

        assert contentids is not None
        output__ = {
            'aggr_stats': aggr_stats,
            'top_model_param': top_model_param,
            'top_ratio': float(count) / len(model_params),
            'statss': statss,
            'model_params': model_params,
            'contentids': contentids,
        }

        return output__

    @classmethod
    def _assert_grid_search(cls, model_param_search_range):
        assert isinstance(model_param_search_range, dict)
        # for grid search, model_param_search_range's values must all be lists
        # or tuples
        for v in model_param_search_range.values():
            assert isinstance(v, (list, tuple))

    @classmethod
    def _assert_random_search(cls, model_param_search_range):
        assert isinstance(model_param_search_range, dict)
        # for random search, model_param_search_range's values must either be
        # lists/tuples, or dictionary containing 'low' and 'high' bounds.
        for v in model_param_search_range.values():
            assert (isinstance(v, (list, tuple))
                    or
                    (isinstance(v, dict) and 'low' in v
                     and 'high' in v and 'decimal' in v)
                    )

    @classmethod
    def print_output(cls, output):
        if 'stats' in output:
            print('Stats: {}'.format(cls.format_stats(output['stats'])))
        if 'aggr_stats' in output:
            print('Aggregated stats: {}'.format(cls.format_stats(output['aggr_stats'])))
        if 'top_model_param' in output:
            print('Top model param ({ratio:.3f}): {modelparam}'.format(
                ratio=output['top_ratio'],
                modelparam=output['top_model_param']))
        if 'statss' in output and 'model_params' in output:
            for fold, (stats, model_param) in \
                    enumerate(zip(output['statss'], output['model_params'])):
                print('Fold {fold}: {model_param}, {stats}'.format(
                    fold=fold, model_param=model_param,
                    stats=cls.format_stats(stats)))

    @staticmethod
    def format_stats(stats):
        return '(SRCC: {srcc:.3f}, PCC: {pcc:.3f}, MSE: {rmse:.3f})'.format(
            srcc=stats['SRCC'], pcc=stats['PCC'], rmse=stats['RMSE'])

    @staticmethod
    def _sample_model_param_list(model_param_search_range, random_search_times):
        keys = sorted(model_param_search_range.keys())  # normalize order
        list_of_dicts = []
        for i in range(random_search_times):
            d = {}
            for k in keys:
                v = model_param_search_range[k]
                if isinstance(v, (list, tuple)):
                    d[k] = random.choice(v)
                elif isinstance(v, dict) and 'low' in v and 'high' in v and 'decimal' in v:
                    num = random.uniform(v['low'], v['high'])
                    scale = (10**v['decimal'])
                    num = int(num * scale) / float(scale)
                    d[k] = num
                else:
                    assert False
            list_of_dicts.append(d)

        return list_of_dicts

    @staticmethod
    def _find_most_frequent_dict(dicts):
        """
        Find dict that appears the most frequently. The issue is to deal with
        that a dictionary is non-hashable. Workaround is to define a hash
        function.
        :param dicts:
        :return:
        """

        def _hash_dict(dict_to_hash):
            return tuple(sorted(dict_to_hash.items()))

        dict_count = {}  # key: hash, value: list of indices
        for idx_dict, input_dict in enumerate(dicts):
            dict_hash = _hash_dict(input_dict)
            dict_count.setdefault(dict_hash, []).append(idx_dict)

        most_frequent_dict_hash = None
        most_frequent_dict_count = None
        for dict_hash in dict_count:
            curr_count = len(dict_count[dict_hash])
            if most_frequent_dict_count is None or most_frequent_dict_count < curr_count:
                most_frequent_dict_hash = dict_hash
                most_frequent_dict_count = curr_count

        # find the dict matching the hash
        most_frequent_dict = None
        for input_dict in dicts:
            if _hash_dict(input_dict) == most_frequent_dict_hash:
                most_frequent_dict = input_dict
                break

        assert most_frequent_dict is not None

        return most_frequent_dict, most_frequent_dict_count



"""
BD-rate calculator. Implementation validated against JCTVC-E137.
"""

from __future__ import annotations

__copyright__ = "Copyright 2016-2024, Netflix, Inc."
__license__ = "BSD+Patent"

import math

import numpy as np
from typing import Iterable, Any

from scipy.integrate import trapezoid
from scipy.interpolate import pchip_interpolate  # type: ignore[attr-defined]

from vmaf.tools.convex_hull import calculate_convex_hull
from .exceptions import (
    BdRateNoOverlapException,
    BdRateNonMonotonicException,
    BdRateZeroRateException,
    BdRateNotEnoughPointsException,
)
from .typing_utils import RdPoint

INF_REPLACEMENT = 100.0
NUM_SAMPLES = 100


def calculate_bd_rate(
    metric_set1: Iterable[RdPoint],
    metric_set2: Iterable[RdPoint],
    min_metric: float | None = None,
    max_metric: float | None = None,
    use_convex_hull: bool = False,
    at_perc: float | None = None,
) -> float:
    """
    Calculate the Bjontegaard Delta rate (BD-Rate).

    Bjontegaard's metric calculates the average percentage saving in bitrate
    between two rate-distortion curves. This implementation uses Piecewise
    Cubic Hermite Interpolating Polynomial (PCHIP) and the trapezoid method
    for integration.

    Args:
       metric_set1: Iterable of tuples (bitrate, metric) for the first graph.
       metric_set2: Iterable of tuples (bitrate, metric) for the second graph.
       min_metric: Optional minimum metric value for integration.
       max_metric: Optional maximum metric value for integration.
       use_convex_hull: If set, the BD-rate computation will be performed on the convex hull of the points.
       at_perc: If set, the BD-rate will be calculated only at a percentage [0.0 - 100.0] where the metric range of the two curves overlaps.

    Returns:
       The average savings in bitrate at equal quality. Multiply by 100 to get a percentage.

    Raises:
       BdRateNotEnoughPointsException: If either metric set has fewer than 4 points.
       BdRateNonMonotonicException: If either curve is non-monotonic.
       BdRateZeroRateException: If any rate in the metric sets is zero.
       BdRateNoOverlapException: If there is no overlapping interval for integration.
    """

    if not metric_set1 or not metric_set2:
        raise BdRateNotEnoughPointsException("One or both of the metric sets is empty or null.")

    if use_convex_hull:
        metric_set1 = calculate_convex_hull(metric_set1)
        metric_set2 = calculate_convex_hull(metric_set2)

    if at_perc is not None:
        if at_perc < 0 or at_perc > 100:
            raise ValueError(f"at_perc must be between 0 and 100, but got {at_perc}.")

    # pchip_interpolate requires keys sorted by x axis.
    # x-axis will be our metric, not the bitrate, so sort by metric.
    metric_set1 = sorted(metric_set1, key=lambda p: p.metric)
    metric_set2 = sorted(metric_set2, key=lambda p: p.metric)

    if len(metric_set1) < 4 or len(metric_set2) < 4:
        raise BdRateNotEnoughPointsException("Each metric set must contain at least 4 points.")

    if not _is_curve_monotonic(metric_set1) or not _is_curve_monotonic(metric_set2):
        raise BdRateNonMonotonicException("One or both curves are non-monotonic.")

    if not _rates_are_nonzero(metric_set1) or not _rates_are_nonzero(metric_set2):
        raise BdRateZeroRateException("One or both metric sets contain zero rates.")

    # Pull the log of the rate and clamped metric from metric_sets.
    log_rate1 = [math.log(x.rate) for x in metric_set1]
    metric1 = [INF_REPLACEMENT if x.metric == float("inf") else x.metric for x in metric_set1]
    log_rate2 = [math.log(x.rate) for x in metric_set2]
    metric2 = [INF_REPLACEMENT if x.metric == float("inf") else x.metric for x in metric_set2]

    # Integration interval. This metric only works on the area that's
    # overlapping. Extrapolation of these things is sketchy so we avoid.
    min_int = max(min(metric1), min(metric2))
    if min_metric:
        min_int = max(min_int, min_metric)
    max_int = min(max(metric1), max(metric2))
    if max_metric:
        max_int = min(max_int, max_metric)

    # No overlap means no sensible metric possible.
    if max_int <= min_int:
        raise BdRateNoOverlapException()

    if at_perc is None:
        # Use Piecewise Cubic Hermite Interpolating Polynomial interpolation to
        # create 100 new samples points separated by interval.
        samples, interval = np.linspace(min_int, max_int, num=NUM_SAMPLES, retstep=True)
        v1 = pchip_interpolate(metric1, log_rate1, samples)
        v2 = pchip_interpolate(metric2, log_rate2, samples)

        # Calculate the integral using the trapezoid method on the samples.
        int_v1 = trapezoid(v1, dx=float(interval))
        int_v2 = trapezoid(v2, dx=float(interval))

        # Calculate the average improvement.
        avg_exp_diff = (int_v2 - int_v1) / (max_int - min_int)

        # Exponentiate to undo the logarithms
        return math.exp(avg_exp_diff) - 1

    else:
        at_metric = min_int + (max_int - min_int) * at_perc / 100.0
        v1b: Any = pchip_interpolate(metric1, log_rate1, [at_metric])
        v2b: Any = pchip_interpolate(metric2, log_rate2, [at_metric])

        # Exponentiate to undo the logarithms
        return math.exp(v2b[0] - v1b[0]) - 1


def _is_curve_monotonic(points: list[RdPoint]) -> bool:
    """
    Check if the RD curve is monotonic.

    Args:
        points: List of RD points.

    Returns:
        True if the curve is monotonic, False otherwise.
    """
    return all(point1.rate < point2.rate and point1.metric < point2.metric for point1, point2 in zip(points, points[1:]))


def _rates_are_nonzero(points: Iterable[RdPoint]) -> bool:
    """
    Check if all rates in the RD points are non-zero.

    Args:
        points: List of RD points.

    Returns:
        True if all rates are non-zero, False otherwise.
    """
    return all(point.rate != 0 for point in points)



from collections.abc import Hashable
from functools import partial
import os

__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"

import warnings
import json
import hashlib
import sys


def deprecated(func):
    """
    Mark a function as deprecated.
    It will result in a warning being emitted when the function is used.
    """

    def new_func(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning) # turn off filter
        warnings.warn("Call to deprecated function {}.".format(func.__name__),
                      category=DeprecationWarning, stacklevel=2)
        warnings.simplefilter('default', DeprecationWarning) # reset filter
        return func(*args, **kwargs)

    new_func.__name__ = func.__name__
    new_func.__doc__ = func.__doc__
    new_func.__dict__.update(func.__dict__)
    return new_func


def persist(original_func):
    """
    Cache returned value of function in a function. Useful when calling functions
    recursively, especially in dynamic programming where lots of returned values
    can be reused.
    """

    cache = {}

    def new_func(*args):
        h = hashlib.sha1(str(original_func.__name__) + str(args)).hexdigest()
        if h not in cache:
            cache[h] = original_func(*args)
        return cache[h]

    return new_func


def dummy(func):
    """ Dummy decorator. """
    return func


class memoized(object):
    """ Decorator. Caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned
    (not reevaluated).

    memoized is similar to persist, but if applied to
    class methods, persist will cache on a per-class basis, while memoized
    will cache on a per-object basis.

    Taken from: https://wiki.python.org/moin/PythonDecoratorLibrary#Memoize
    """

    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args):
        if not isinstance(args, Hashable):
            return self.func(*args)
        if args in self.cache:
            return self.cache[args]
        else:
            value = self.func(*args)
            self.cache[args] = value
            return value

    def __repr__(self):
        """ Return the function's docstring. """
        return self.func.__doc__

    def __get__(self, obj, objtype):
        """ Support instance methods. """
        return partial(self.__call__, obj)


def persist_to_file(file_name):
    """
    Cache (or persist) returned value of function in a json file .
    """

    def decorator(original_func):

        if not os.path.exists(file_name):
            cache = {}
        else:
            try:
                cache = json.load(open(file_name, 'rt'))
            except (IOError, ValueError):
                sys.exit(1)

        def new_func(*args):
            h = hashlib.sha1(str(original_func.__name__) + str(args)).hexdigest()
            if h not in cache:
                cache[h] = original_func(*args)
                file_dir = os.path.dirname(file_name)
                os.makedirs(file_dir, exist_ok=True)
                json.dump(cache, open(file_name, 'wt'))
            return cache[h]

        return new_func

    return decorator


def persist_to_dir(dir_name):
    """
    Cache (or persist) returned value of function in a directory of files.
    """

    def decorator(original_func):

        def new_func(*args):
            h = hashlib.sha1(str(original_func.__name__) + str(args)).hexdigest()
            file_name = os.path.join(dir_name, h)
            if not os.path.exists(file_name):
                os.makedirs(dir_name, exist_ok=True)
                res = original_func(*args)
                json.dump(res, open(file_name, 'wt'))
            else:
                res = json.load(open(file_name, 'rt'))
            return res

        return new_func

    return decorator


def override(interface_class):
    def overrider(method):
        assert(method.__name__ in dir(interface_class)), \
            f"{method.__name__} does not override any method in {interface_class.__name__}"
        return method
    return overrider


class change_repr(object):
    def __init__(self, functor):
        self.functor = functor
        #  lets copy some key attributes from the original function
        self.__name__ = functor.__name__
        self.__doc__ = functor.__doc__

    def __call__(self, *args, **kwargs):
        return self.functor(*args, **kwargs)

    def __repr__(self):
        return self.functor.__name__



feature_dict = {

    # use selected features from VmafFeatureExtractor
    'VMAF_feature': ['vif', 'adm', 'motion', 'ansnr'],

    'Moment_feature':['dis1st'], # use distorted video's 1st moment
}

feature_optional_dict = {
    'VMAF_feature': {'adm_ref_display_height': 108000},
}



import shutil
from abc import ABCMeta, abstractmethod
import multiprocessing
import os
from time import sleep
import hashlib
from typing import Optional, List, Tuple

from vmaf.core.asset import Asset
from vmaf.tools.decorator import deprecated, override

from vmaf.tools.misc import make_parent_dirs_if_nonexist, get_dir_without_last_slash, \
    parallel_map, match_any_files, run_process, \
    get_file_name_extension
from vmaf.core.mixin import TypeVersionEnabled
from vmaf.config import VmafExternalConfig
from vmaf.tools.reader import YuvReader
from vmaf.tools.writer import YuvWriter

__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"


class Executor(TypeVersionEnabled):
    """
    An Executor takes in a list of Assets, and run computations on them, and
    return a list of corresponding Results. An Executor must specify a unique
    type and version combination (by the TYPE and VERSION attribute), so that
    the Result generated by it can be uniquely identified.

    Executor is the base class for FeatureExtractor and QualityRunner, and it
    provides a number of shared housekeeping functions, including reusing
    Results, creating FIFO pipes, cleaning up log files/Results, etc.

    (3/11/2020) added an (optional) step to allow python-based processing on
    both the ref and dis files. The new processing pipeline looks like this:

     notyuv  --------   ref_workfile   -----------------   ref_procfile
    -------> |FFmpeg| ---------------> |python-callback| -----------------
             --------                  -----------------                 |    -----------------
                                                                         ---> |               |
                                                                              |     VMAF      | --->
                                                                         ---> |               |
     notyuv  --------   dis_workfile   -----------------   dis_procfile  |    -----------------
    -------> |FFmpeg| ---------------> |python-callback| ----------------
             --------                  -----------------
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def _generate_result(self, asset):
        raise NotImplementedError

    @abstractmethod
    def _read_result(self, asset):
        raise NotImplementedError

    def __init__(self,
                 assets,
                 logger,
                 fifo_mode=True,
                 delete_workdir=True,
                 result_store=None,
                 optional_dict=None,
                 optional_dict2=None,
                 save_workfiles=False,
                 ):
        """
        Use optional_dict for parameters that would impact result (e.g. model,
        patch size), and use optional_dict2 for parameters that would NOT
        impact result (e.g. path to data cache file).
        """

        TypeVersionEnabled.__init__(self)

        self.assets = assets
        self.logger = logger
        self.fifo_mode = fifo_mode
        self.delete_workdir = delete_workdir
        self.results = []
        self.result_store = result_store
        self.optional_dict = optional_dict
        self.optional_dict2 = optional_dict2
        self.save_workfiles = save_workfiles

        self._assert_class()
        self._assert_args()
        self._assert_assets()

        self._custom_init()

    def _custom_init(self):
        pass

    @staticmethod
    def get_normalized_string_from_dict(d):
        """ Normalized string representation with sorted keys, extended to cover callables.

        >>> Executor.get_normalized_string_from_dict({"max_buffer_sec": 5.0, "bitrate_kbps": 45, })
        'bitrate_kbps_45_max_buffer_sec_5.0'
        >>> Executor.get_normalized_string_from_dict({"sorted_key": sorted, "_need_ffmpeg_key": Executor._need_ffmpeg, })
        '_need_ffmpeg_key_Executor._need_ffmpeg_sorted_key_sorted'
        """
        def _slugify(v):
            if callable(v):
                s = str(v)
                assert s[0] == '<' and s[-1] == '>'
                s = s[1:-1]
                l = s.split(' ')
                assert 'function' in l
                for idx, e in enumerate(l):
                    if e == 'function':
                        assert idx < len(l) - 1
                        return l[idx + 1]
            else:
                return v

        normalized_str = '_'.join(map(lambda k: '{k}_{v}'.format(k=k, v=_slugify(d[k])), sorted(d.keys())))

        if len(normalized_str) > 140:  # upper limit of filename is 256 but leave some space for prefix/suffix
            normalized_str = hashlib.sha1(normalized_str.encode("utf-8")).hexdigest()

        return normalized_str

    @property
    def executor_id(self):
        executor_id_ = TypeVersionEnabled.get_type_version_string(self)

        if self.optional_dict is not None and len(self.optional_dict) > 0:
            # include optional_dict info in executor_id for result store,
            # as parameters in optional_dict will impact result
            executor_id_ += '_{}'.format(self.get_normalized_string_from_dict(self.optional_dict))
            replace_chars = ["'", " "]
            for c in replace_chars:
                executor_id_ = executor_id_.replace(c, "_")
        return executor_id_

    def run(self, **kwargs):
        """
        Do all the computation here.
        :return:
        """
        if self.logger:
            self.logger.info(
                "For each asset, if {type} result has not been generated, run "
                "and generate {type} result...".format(type=self.executor_id))

        if 'parallelize' in kwargs:
            parallelize = kwargs['parallelize']
        else:
            parallelize = False
        assert isinstance(parallelize, bool)

        if 'processes' in kwargs and kwargs['processes'] is not None:
            assert parallelize is True, 'Cannot specify processes if parallelize is False.'
            processes = kwargs['processes']
        else:
            processes = None
        assert processes is None or (isinstance(processes, int) and processes >= 1)

        if parallelize:
            # create locks for unique assets (uniqueness is identified by str(asset))
            map_asset_lock = {}
            locks = []
            for asset in self.assets:
                asset_str = str(asset)
                if asset_str not in map_asset_lock:
                    map_asset_lock[asset_str] = multiprocessing.Lock()
                locks.append(map_asset_lock[asset_str])

            # pack key arguments to be used as inputs to map function
            list_args = []
            for asset, lock in zip(self.assets, locks):
                list_args.append(
                    [asset, lock])

            def _run(asset_lock):
                asset, lock = asset_lock
                lock.acquire()
                result = self._run_on_asset(asset)
                lock.release()
                return result

            self.results = parallel_map(_run, list_args, processes=processes)
        else:
            self.results = list(map(self._run_on_asset, self.assets))

    def remove_results(self):
        """
        Remove all relevant Results stored in ResultStore, which is specified
        at the constructor.
        :return:
        """
        for asset in self.assets:
            self._remove_result(asset)

    @classmethod
    def _assert_class(cls):
        pass

    def _assert_args(self):
        if self.save_workfiles is True:
            assert self.fifo_mode is False, 'To save workfiles, FIFO mode cannot be true.'

    def _assert_assets(self):

        for asset in self.assets:
            self._assert_an_asset(asset)

        pass

    @staticmethod
    def _need_ffmpeg(asset):
        # 1) if quality width/height do not agree with ref/dis width/height,
        # must rely on ffmpeg for scaling
        # 2) if crop/pad/etc. is needed, need ffmpeg
        # 3) if ref/dis videos' start/end frames specified, need ffmpeg for
        # frame extraction
        # 4) if workfile_yuv_type specified and ref/yuv's yuv_type does not agree,
        # need ffmpeg for conversion
        ret = asset.quality_width_height != asset.ref_width_height \
               or asset.quality_width_height != asset.dis_width_height \
               or asset.ref_yuv_type == 'notyuv' or asset.dis_yuv_type == 'notyuv' \
               or asset.ref_start_end_frame is not None or asset.dis_start_end_frame is not None \
               or 'workfile_yuv_type' in asset.asset_dict and (asset.workfile_yuv_type != asset.ref_yuv_type or asset.workfile_yuv_type != asset.dis_yuv_type)
        for key in Asset.ORDERED_FILTER_LIST:
            ret = ret or asset.get_filter_cmd(key, 'ref') is not None or asset.get_filter_cmd(key, 'dis') is not None

        return ret

    @classmethod
    def _assert_an_asset(cls, asset):

        # needed by _generate_result, and by _open_ref_workfile or
        # _open_dis_workfile if called
        assert asset.quality_width_height is not None

        if cls._need_ffmpeg(asset):
            VmafExternalConfig.get_and_assert_ffmpeg()

        # ref_yuv_type and dis_yuv_type must match, unless any of them is notyuv.
        # also check the logic in _get_workfile_yuv_type
        assert (asset.ref_yuv_type == 'notyuv' or asset.dis_yuv_type == 'notyuv') \
               or (asset.ref_yuv_type == asset.dis_yuv_type) or "workfile_yuv_type" in asset.asset_dict

        # if crop_cmd or pad_cmd or etc. is specified, make sure quality_width and
        # quality_height are EXPLICITLY specified in asset_dict
        if asset.ref_crop_cmd is not None or asset.dis_crop_cmd is not None:
            assert 'quality_width' in asset.asset_dict and 'quality_height' in asset.asset_dict, \
                'If crop_cmd or etc. is specified, must also EXPLICITLY specify quality_width and quality_height.'
        if asset.ref_pad_cmd is not None or asset.dis_pad_cmd is not None:
            assert 'quality_width' in asset.asset_dict and 'quality_height' in asset.asset_dict, \
                'If pad_cmd or etc. is specified, must also EXPLICITLY specify quality_width and quality_height.'

    @staticmethod
    def _get_workfile_yuv_type(asset):
        """ Same as original yuv type, unless it is notyuv; in this case, check
        the other's (if ref, check dis'; vice versa); if both notyuv, use format as set at a higher level"""

        # also check the logic in _assert_an_asset. The assumption is:
        # assert (asset.ref_yuv_type == 'notyuv' or asset.dis_yuv_type == 'notyuv') \
        #        or (asset.ref_yuv_type == asset.dis_yuv_type)

        # if the workfile_yuv_type is provided by the user, then we use it
        if "workfile_yuv_type" in asset.asset_dict:
            return asset.workfile_yuv_type

        if asset.ref_yuv_type == 'notyuv' and asset.dis_yuv_type == 'notyuv':
            return asset.workfile_yuv_type
        elif asset.ref_yuv_type == 'notyuv' and asset.dis_yuv_type != 'notyuv':
            return asset.dis_yuv_type
        elif asset.ref_yuv_type != 'notyuv' and asset.dis_yuv_type == 'notyuv':
            return asset.ref_yuv_type
        else:  # neither notyuv
            assert asset.ref_yuv_type == asset.dis_yuv_type, "YUV types for ref and dis do not match."
            return asset.ref_yuv_type

    def _wait_for_workfiles(self, asset):
        # wait til workfile paths being generated
        for i in range(10):
            if os.path.exists(asset.ref_workfile_path) and os.path.exists(asset.dis_workfile_path):
                break
            sleep(0.1)
        else:
            raise RuntimeError("ref or dis video workfile path {ref} or {dis} is missing.".format(ref=asset.ref_workfile_path, dis=asset.dis_workfile_path))

    def _wait_for_procfiles(self, asset):
        # wait til procfile paths being generated
        for i in range(10):
            if os.path.exists(asset.ref_procfile_path) and os.path.exists(asset.dis_procfile_path):
                break
            sleep(0.1)
        else:
            raise RuntimeError("ref or dis video procfile path {ref} or {dis} is missing.".format(ref=asset.ref_procfile_path, dis=asset.dis_procfile_path))

    def _prepare_log_file(self, asset):

        log_file_path = self._get_log_file_path(asset)

        # if parent dir doesn't exist, create
        make_parent_dirs_if_nonexist(log_file_path)

        # add runner type and version
        with open(log_file_path, 'wt') as log_file:
            log_file.write("{type_version_str}\n\n".format(
                type_version_str=self.get_cozy_type_version_string()))

    def _assert_paths(self, asset):
        assert os.path.exists(asset.ref_path) or match_any_files(asset.ref_path), \
            "Reference path {} does not exist.".format(asset.ref_path)
        assert os.path.exists(asset.dis_path) or match_any_files(asset.dis_path), \
            "Distorted path {} does not exist.".format(asset.dis_path)

    def _run_on_asset(self, asset):
        # Wraper around the essential function _generate_result, to
        # do housekeeping work including 1) asserts of asset, 2) skip run if
        # log already exist, 3) creating fifo, 4) delete work file and dir

        if self.result_store:
            result = self.result_store.load(asset, self.executor_id)
            qw, qh = asset.quality_width_height
            if result is not None and self.save_workfiles is True \
                    and not self.result_store.has_workfile(asset, self.executor_id, f'_dis.{qw}x{qh}.{self._get_workfile_yuv_type(asset)}.yuv'):
                result = None  # if save_workfiles is True and has_workfile is False, invalidate result and rerun
        else:
            result = None

        # if result can be retrieved from result_store, skip log file
        # generation and reading result from log file, but directly
        # return the retrieved result
        if result is not None:
            if self.logger:
                self.logger.info('{id} result exists. Skip {id} run.'.
                                 format(id=self.executor_id))
        else:

            if self.logger:
                self.logger.info('{id} result does\'t exist. Perform {id} '
                                 'calculation.'.format(id=self.executor_id))

            # at this stage, it is certain that asset.ref_path and
            # asset.dis_path will be used. must early determine that
            # they exists
            self._assert_paths(asset)

            # if no FFmpeg is involved, directly work on ref_path/dis_path,
            # instead of opening workfiles
            self._set_asset_use_path_as_workpath(asset)

            # if no ref/dis_proc_callback is involved, directly work on ref/dis_workfile_path,
            # instead of opening procfiles
            self._set_asset_use_workpath_as_procpath(asset)

            # remove workfiles if exist (do early here to avoid race condition
            # when ref path and dis path have some overlap)
            if asset.use_path_as_workpath:
                # do nothing
                pass
            else:
                self._close_workfiles(asset)

            # remove procfiles if exist (do early here to avoid race condition
            # when ref path and dis path have some overlap)
            if asset.use_workpath_as_procpath:
                # do nothing
                pass
            else:
                self._close_procfiles(asset)

            log_file_path = self._get_log_file_path(asset)
            make_parent_dirs_if_nonexist(log_file_path)

            if asset.use_path_as_workpath:
                # do nothing
                pass
            else:
                if self.fifo_mode:
                    self._open_workfiles_in_fifo_mode(asset)
                else:
                    self._open_workfiles(asset)

            if asset.use_workpath_as_procpath:
                # do nothing
                pass
            else:
                if self.fifo_mode:
                    self._open_procfiles_in_fifo_mode(asset)
                else:
                    self._open_procfiles(asset)

            self._prepare_log_file(asset)

            self._generate_result(asset)

            if self.logger:
                self.logger.info("Read {id} log file, get scores...".
                                 format(id=self.executor_id))

            # collect result from each asset's log file
            result = self._read_result(asset)

            # save result
            if self.result_store:
                result = self._save_result(result)

            # clean up workfiles
            if self.delete_workdir:
                if asset.use_path_as_workpath:
                    # do nothing
                    pass
                else:
                    self._close_workfiles(asset)

                if asset.use_workpath_as_procpath:
                    # do nothing
                    pass
                else:
                    self._close_procfiles(asset)

            # clean up workdir and log files in it
            if self.delete_workdir:

                # remove log file
                self._remove_log(asset)

                # remove dir
                log_file_path = self._get_log_file_path(asset)
                log_dir = get_dir_without_last_slash(log_file_path)
                shutil.rmtree(log_dir)

        result = self._post_process_result(result)

        return result

    def _open_workfiles(self, asset):
        self._open_ref_workfile(asset, fifo_mode=False)
        self._open_dis_workfile(asset, fifo_mode=False)

    def _open_workfiles_in_fifo_mode(self, asset):
        ref_p = multiprocessing.Process(target=self._open_ref_workfile,
                                        args=(asset, True))
        dis_p = multiprocessing.Process(target=self._open_dis_workfile,
                                        args=(asset, True))
        ref_p.start()
        dis_p.start()
        self._wait_for_workfiles(asset)

    def _open_procfiles(self, asset):
        self._open_ref_procfile(asset, fifo_mode=False)
        self._open_dis_procfile(asset, fifo_mode=False)

    def _open_procfiles_in_fifo_mode(self, asset):
        ref_p = multiprocessing.Process(target=self._open_ref_procfile,
                                        args=(asset, True))
        dis_p = multiprocessing.Process(target=self._open_dis_procfile,
                                        args=(asset, True))
        ref_p.start()
        dis_p.start()
        self._wait_for_procfiles(asset)

    def _close_workfiles(self, asset):
        self._close_ref_workfile(asset)
        self._close_dis_workfile(asset)

    @classmethod
    def _close_procfiles(cls, asset):
        cls._close_ref_procfile(asset)
        cls._close_dis_procfile(asset)

    def _refresh_workfiles_before_additional_pass(self, asset):
        # If fifo mode and workpath needs to be freshly generated, must
        # reopen the fifo pipe before proceeding
        if self.fifo_mode and (not asset.use_path_as_workpath):
            self._close_workfiles(asset)
            self._open_workfiles_in_fifo_mode(asset)

    def _save_result(self, result):
        self.result_store.save(result)
        if self.save_workfiles:
            qw, qh = result.asset.quality_width_height
            self.result_store.save_workfile(result, result.asset.ref_workfile_path, f'_ref.{qw}x{qh}.{self._get_workfile_yuv_type(result.asset)}.yuv')
            self.result_store.save_workfile(result, result.asset.dis_workfile_path, f'_dis.{qw}x{qh}.{self._get_workfile_yuv_type(result.asset)}.yuv')
        return result

    @classmethod
    def _set_asset_use_path_as_workpath(cls, asset):
        # if no rescaling or croping or padding or etc. is involved, directly work on
        # ref_path/dis_path, instead of opening workfiles
        if not cls._need_ffmpeg(asset):
            asset.use_path_as_workpath = True

    @classmethod
    def _set_asset_use_workpath_as_procpath(cls, asset):
        if asset.ref_proc_callback is None and asset.dis_proc_callback is None:
            asset.use_workpath_as_procpath = True

    @classmethod
    def _post_process_result(cls, result):
        # do nothing, wait to be overridden
        return result

    def _get_log_file_path(self, asset):
        return "{workdir}/{executor_id}_{str}".format(
            workdir=asset.workdir, executor_id=self.executor_id,
            str=hashlib.sha1(str(asset).encode("utf-8")).hexdigest())

    # ===== workfile =====

    def _open_ref_workfile(self, asset, fifo_mode):

        use_path_as_workpath = asset.use_path_as_workpath
        path = asset.ref_path
        workfile_path = asset.ref_workfile_path
        quality_width_height = self._get_quality_width_height(asset)
        yuv_type = asset.ref_yuv_type
        try:
            decoder_type = asset.ref_decoder_type
        except AttributeError:
            decoder_type = None
        try:
            preresampling_filterchain = asset.ref_preresampling_filterchain
        except AttributeError:
            preresampling_filterchain = None
        resampling_type = self._get_ref_resampling_type(asset)
        try:
            postresampling_filterchain = asset.ref_postresampling_filterchain
        except AttributeError:
            postresampling_filterchain = None
        width_height = asset.ref_width_height
        ref_or_dis = 'ref'
        workfile_yuv_type = self._get_workfile_yuv_type(asset)
        logger = self.logger

        _open_workfile_method = self.optional_dict2['_open_workfile_method'] \
            if self.optional_dict2 is not None and '_open_workfile_method' in self.optional_dict2 and self.optional_dict2['_open_workfile_method'] is not None \
            else self._open_workfile
        _open_workfile_method(self, asset, path, workfile_path, yuv_type, workfile_yuv_type, decoder_type,
                              preresampling_filterchain, resampling_type, postresampling_filterchain,
                              width_height, quality_width_height, ref_or_dis, use_path_as_workpath, fifo_mode, logger)

    def _open_dis_workfile(self, asset, fifo_mode):

        use_path_as_workpath = asset.use_path_as_workpath
        path = asset.dis_path
        workfile_path = asset.dis_workfile_path
        quality_width_height = self._get_quality_width_height(asset)
        yuv_type = asset.dis_yuv_type
        try:
            decoder_type = asset.dis_decoder_type
        except AttributeError:
            decoder_type = None
        try:
            preresampling_filterchain = asset.dis_preresampling_filterchain
        except AttributeError:
            preresampling_filterchain = None
        resampling_type = self._get_dis_resampling_type(asset)
        try:
            postresampling_filterchain = asset.dis_postresampling_filterchain
        except AttributeError:
            postresampling_filterchain = None
        width_height = asset.dis_width_height
        ref_or_dis = 'dis'
        workfile_yuv_type = self._get_workfile_yuv_type(asset)
        logger = self.logger

        _open_workfile_method = self.optional_dict2['_open_workfile_method'] \
            if self.optional_dict2 is not None and '_open_workfile_method' in self.optional_dict2 and self.optional_dict2['_open_workfile_method'] is not None \
            else self._open_workfile
        _open_workfile_method(self, asset, path, workfile_path, yuv_type, workfile_yuv_type, decoder_type,
                              preresampling_filterchain, resampling_type, postresampling_filterchain,
                              width_height, quality_width_height, ref_or_dis, use_path_as_workpath, fifo_mode, logger)

    @staticmethod
    def _open_workfile(cls, asset, path, workfile_path, yuv_type, workfile_yuv_type, decoder_type: Optional[str],
                       preresampling_filterchain: Optional[List[str]], resampling_type: str, postresampling_filterchain: Optional[List[str]],
                       width_height: Optional[Tuple[int, int]], quality_width_height: Tuple[int, int], ref_or_dis, use_path_as_workpath, fifo_mode, logger):

        # decoder type must be None here
        assert decoder_type is None, f'decoder_type must be None but is: {decoder_type}'

        # preresampling_filterchain must be None here
        assert preresampling_filterchain is None, f'preresampling_filterchain mut be None but is: {preresampling_filterchain}'

        # postresampling_filterchain msut be None here
        assert postresampling_filterchain is None, f'postresampling_filterchain must be None but is: {postresampling_filterchain}'

        # only need to open workfile if the path is different from path
        assert use_path_as_workpath is False and path != workfile_path
        # if fifo mode, mkfifo
        if fifo_mode:
            os.mkfifo(workfile_path)

        if ref_or_dis == 'ref':
            start_end_frame = asset.ref_start_end_frame
        elif ref_or_dis == 'dis':
            start_end_frame = asset.dis_start_end_frame
        else:
            assert False

        if yuv_type != 'notyuv':
            # in this case, for sure has width_height
            assert width_height is not None
            width, height = width_height
            src_fmt_cmd = cls._get_yuv_src_fmt_cmd(yuv_type, height, width)
        else:
            src_fmt_cmd = cls._get_notyuv_src_fmt_cmd(path)

        vframes_cmd, select_cmd = cls._get_vframes_cmd(start_end_frame)

        crop_cmd = cls._get_filter_cmd(asset, 'crop', ref_or_dis)
        pad_cmd = cls._get_filter_cmd(asset, 'pad', ref_or_dis)
        quality_width, quality_height = quality_width_height
        scale_cmd = f'scale={quality_width}x{quality_height}'
        filter_cmds = []
        for key in Asset.ORDERED_FILTER_LIST:
            if key != 'crop' and key != 'pad':
                filter_cmds.append(cls._get_filter_cmd(asset, key, ref_or_dis))
        vf_cmd = [select_cmd, crop_cmd, pad_cmd, scale_cmd] + filter_cmds
        vf_cmd = ','.join(filter(lambda s: s != '', vf_cmd))

        ffmpeg_cmd = []
        ffmpeg_cmd += [VmafExternalConfig.get_and_assert_ffmpeg()]
        ffmpeg_cmd += [src_fmt_cmd]
        ffmpeg_cmd += ['-i', path]
        ffmpeg_cmd += ['-an', '-vsync', '0']
        ffmpeg_cmd += ['-pix_fmt', workfile_yuv_type]
        ffmpeg_cmd += [vframes_cmd]
        ffmpeg_cmd += ['-vf', vf_cmd]
        ffmpeg_cmd += ['-f', 'rawvideo']
        ffmpeg_cmd += ['-sws_flags', resampling_type]
        ffmpeg_cmd += ['-y', '-nostdin']
        ffmpeg_cmd += [workfile_path]
        ffmpeg_cmd = ' '.join(ffmpeg_cmd)
        if logger:
            logger.info(ffmpeg_cmd)
        run_process(ffmpeg_cmd, shell=True, env=VmafExternalConfig.ffmpeg_env())

    # ===== procfile =====

    def _open_ref_procfile(self, asset, fifo_mode):

        # only need to open ref procfile if the path is different from ref path
        assert asset.use_workpath_as_procpath is False and asset.ref_workfile_path != asset.ref_procfile_path

        ref_proc_callback = asset.ref_proc_callback if asset.ref_proc_callback is not None else lambda x: x

        if fifo_mode:
            os.mkfifo(asset.ref_procfile_path)

        quality_width, quality_height = self._get_quality_width_height(asset)
        yuv_type = asset.workfile_yuv_type
        with YuvReader(filepath=asset.ref_workfile_path, width=quality_width, height=quality_height,
                       yuv_type=yuv_type) as ref_yuv_reader:
            with YuvWriter(filepath=asset.ref_procfile_path, width=quality_width, height=quality_height,
                           yuv_type=yuv_type) as ref_yuv_writer:
                while True:
                    try:
                        y, u, v = ref_yuv_reader.next(format='float')
                        y, u, v = ref_proc_callback(y), u, v
                        ref_yuv_writer.next(y, u, v, format='float2uint')
                    except StopIteration:
                        break

    def _open_dis_procfile(self, asset, fifo_mode):

        # only need to open dis procfile if the path is different from dis path
        assert asset.use_workpath_as_procpath is False and asset.dis_workfile_path != asset.dis_procfile_path

        dis_proc_callback = asset.dis_proc_callback if asset.dis_proc_callback is not None else lambda x: x

        if fifo_mode:
            os.mkfifo(asset.dis_procfile_path)

        quality_width, quality_height = self._get_quality_width_height(asset)
        yuv_type = asset.workfile_yuv_type
        with YuvReader(filepath=asset.dis_workfile_path, width=quality_width, height=quality_height,
                       yuv_type=yuv_type) as dis_yuv_reader:
            with YuvWriter(filepath=asset.dis_procfile_path, width=quality_width, height=quality_height,
                           yuv_type=yuv_type) as dis_yuv_writer:
                while True:
                    try:
                        y, u, v = dis_yuv_reader.next(format='float')
                        y, u, v = dis_proc_callback(y), u, v
                        dis_yuv_writer.next(y, u, v, format='float2uint')
                    except StopIteration:
                        break

    def _get_ref_resampling_type(self, asset):
        return asset.ref_resampling_type

    def _get_dis_resampling_type(self, asset):
        return asset.dis_resampling_type

    def _get_quality_width_height(self, asset):
        return asset.quality_width_height

    @staticmethod
    def _get_yuv_src_fmt_cmd(yuv_type, height, width):
        yuv_src_fmt_cmd = '-f rawvideo -pix_fmt {yuv_fmt} -s {width}x{height}'. \
            format(yuv_fmt=yuv_type, width=width, height=height)
        return yuv_src_fmt_cmd

    @staticmethod
    def _get_notyuv_src_fmt_cmd(path):
        if get_file_name_extension(path) in ['j2c', 'j2k', 'tiff']:
            # 2147483647 is INT_MAX if int is 4 bytes
            return "-f image2 -start_number_range 2147483647"
        elif get_file_name_extension(path) in ['icpf']:
            return "-f image2 -c:v netflixprores -start_number_range 2147483647"
        elif get_file_name_extension(path) in ['265']:
            return "-c:v hevc"
        else:
            return ""

    @staticmethod
    def _get_filter_cmd(asset, key, target):
        return "{}={}".format(key, asset.get_filter_cmd(key, target)) \
            if asset.get_filter_cmd(key, target) is not None else ""

    @staticmethod
    def _get_vframes_cmd(start_end_frame: Optional[list]):
        if start_end_frame is None:
            return "", ""
        else:
            start_frame, end_frame = start_end_frame
            num_frames = end_frame - start_frame + 1
            return f"-vframes {num_frames}", f"select='gte(n\\,{start_frame})*gte({end_frame}\\,n)',setpts=PTS-STARTPTS"

    def _close_ref_workfile(self, asset):

        use_path_as_workpath = asset.use_path_as_workpath
        path = asset.ref_path
        workfile_path = asset.ref_workfile_path

        _close_workfile_method = self.optional_dict2['_close_workfile_method'] \
            if self.optional_dict2 is not None and '_close_workfile_method' in self.optional_dict2 and self.optional_dict2['_close_workfile_method'] is not None \
            else self._close_workfile
        _close_workfile_method(path, workfile_path, use_path_as_workpath)

    def _close_dis_workfile(self, asset):

        use_path_as_workpath = asset.use_path_as_workpath
        path = asset.dis_path
        workfile_path = asset.dis_workfile_path

        _close_workfile_method = self.optional_dict2['_close_workfile_method'] \
            if self.optional_dict2 is not None and '_close_workfile_method' in self.optional_dict2 and self.optional_dict2['_close_workfile_method'] is not None \
            else self._close_workfile
        _close_workfile_method(path, workfile_path, use_path_as_workpath)

    @staticmethod
    def _close_workfile(path, workfile_path, use_path_as_workpath):

        # only need to close workfile if the workfile path is different from path
        assert use_path_as_workpath is False and path != workfile_path

        if os.path.exists(workfile_path):
            os.remove(workfile_path)

    @staticmethod
    def _close_ref_procfile(asset):

        # only need to close ref procfile if the path is different from ref workpath
        assert asset.use_workpath_as_procpath is False and asset.ref_workfile_path != asset.ref_procfile_path

        if os.path.exists(asset.ref_procfile_path):
            os.remove(asset.ref_procfile_path)

    @staticmethod
    def _close_dis_procfile(asset):

        # only need to close dis procfile if the path is different from dis path
        assert asset.use_workpath_as_procpath is False and asset.dis_workfile_path != asset.dis_procfile_path

        if os.path.exists(asset.dis_procfile_path):
            os.remove(asset.dis_procfile_path)

    def _remove_log(self, asset):
        log_file_path = self._get_log_file_path(asset)
        if os.path.exists(log_file_path):
            os.remove(log_file_path)

    def _remove_result(self, asset):
        if self.result_store:
            self.result_store.delete(asset, self.executor_id)
            qw, qh = asset.quality_width_height
            self.result_store.delete_workfile(asset, self.executor_id, f'_ref.{qw}x{qh}.{self._get_workfile_yuv_type(asset)}.yuv')
            self.result_store.delete_workfile(asset, self.executor_id, f'_dis.{qw}x{qh}.{self._get_workfile_yuv_type(asset)}.yuv')

@deprecated
def run_executors_in_parallel(executor_class,
                              assets,
                              fifo_mode=True,
                              delete_workdir=True,
                              parallelize=True,
                              logger=None,
                              result_store=None,
                              optional_dict=None,
                              optional_dict2=None,
                              ):
    """
    Run multiple Executors in parallel.
    """

    # construct an executor object just to call _assert_assets() only
    executor_class(
        assets,
        logger,
        fifo_mode=fifo_mode,
        delete_workdir=True,
        result_store=result_store,
        optional_dict=optional_dict,
        optional_dict2=optional_dict2
    )

    # create locks for unique assets (uniqueness is identified by str(asset))
    map_asset_lock = {}
    locks = []
    for asset in assets:
        asset_str = str(asset)
        if asset_str not in map_asset_lock:
            map_asset_lock[asset_str] = multiprocessing.Lock()
        locks.append(map_asset_lock[asset_str])

    # pack key arguments to be used as inputs to map function
    list_args = []
    for asset, lock in zip(assets, locks):
        list_args.append(
            [executor_class, asset, fifo_mode, delete_workdir,
             result_store, optional_dict, optional_dict2, lock])

    def run_executor(args):
        executor_class, asset, fifo_mode, delete_workdir, \
        result_store, optional_dict, optional_dict2, lock = args
        lock.acquire()
        executor = executor_class([asset], None, fifo_mode, delete_workdir,
                                  result_store, optional_dict, optional_dict2)
        executor.run()
        lock.release()
        return executor

    # run
    if parallelize:
        executors = parallel_map(run_executor, list_args, processes=None)
    else:
        executors = list(map(run_executor, list_args))

    # aggregate results
    results = [executor.results[0] for executor in executors]

    return executors, results


class NorefExecutorMixin(object):
    """
    Override Executor whenever reference video is mentioned.

    NorefExecutorMixin is useful for NorefAsset, i.e. assets that does not have a reference.

    Example classes that inherits NorefExecutorMixin include: BrisqueNorefFeatureExtractor, VideoEncoder

    (3/11/2020) added an (optional) step to allow python-based processing on
    the dis file. The new processing pipeline looks like this:

     notyuv  --------   dis_workfile   -----------------    dis_procfile      -----------------
    -------> |FFmpeg| ---------------> |python-callback| -------------------> |   BRISQUE     | --->
             --------                  -----------------                      -----------------
    """

    @staticmethod
    @override(Executor)
    def _need_ffmpeg(asset):
        # 1) if quality width/height do not to agree with dis width/height,
        # must rely on ffmpeg for scaling
        # 2) if crop/pad/etc. is need, need ffmpeg
        # 3) if dis videos' start/end frames specified, need ffmpeg for
        # frame extraction
        # 4) if workfile_yuv_type specified and doesn't agree with the dis yuv_type
        ret = asset.quality_width_height != asset.dis_width_height \
               or asset.dis_yuv_type == 'notyuv' \
               or asset.dis_start_end_frame is not None \
               or 'workfile_yuv_type' in asset.asset_dict and asset.workfile_yuv_type != asset.dis_yuv_type
        for key in Asset.ORDERED_FILTER_LIST:
            ret = ret or asset.get_filter_cmd(key, 'dis') is not None

        return ret

    @classmethod
    def _assert_an_asset(cls, asset):

        # needed by _generate_result, and by _open_dis_workfile if called
        assert asset.quality_width_height is not None

        if cls._need_ffmpeg(asset):
            VmafExternalConfig.get_and_assert_ffmpeg()

        # if crop_cmd or pad_cmd or etc. is specified, make sure quality_width and
        # quality_height are EXPLICITLY specified in asset_dict
        if asset.dis_crop_cmd is not None:
            assert 'quality_width' in asset.asset_dict and 'quality_height' in asset.asset_dict, \
                'If crop_cmd etc. is specified, must also EXPLICITLY specify quality_width and quality_height.'
        if asset.dis_pad_cmd is not None:
            assert 'quality_width' in asset.asset_dict and 'quality_height' in asset.asset_dict, \
                'If pad_cmd etc. is specified, must also EXPLICITLY specify quality_width and quality_height.'

    @staticmethod
    def _get_workfile_yuv_type(asset):
        """ Same as original yuv type, unless it is notyuv or specified by the user; 
        in this case, use format as set at a higher level"""

        if 'workfile_yuv_type' in asset.asset_dict or asset.dis_yuv_type == 'notyuv':
            return asset.workfile_yuv_type
        else:
            return asset.dis_yuv_type

    @override(Executor)
    def _wait_for_workfiles(self, asset):
        # wait til workfile paths being generated
        for i in range(10):
            if os.path.exists(asset.dis_workfile_path):
                break
            sleep(0.1)
        else:
            raise RuntimeError("dis video workfile path {} is missing.".format(
                asset.dis_workfile_path))

    @override(Executor)
    def _wait_for_procfiles(self, asset):
        # wait til procfile paths being generated
        for i in range(10):
            if os.path.exists(asset.dis_procfile_path):
                break
            sleep(0.1)
        else:
            raise RuntimeError("dis video procfile path {} is missing.".format(
                asset.dis_procfile_path))

    @override(Executor)
    def _assert_paths(self, asset):
        assert os.path.exists(asset.dis_path) or match_any_files(asset.dis_path), \
            "Distorted path {} does not exist.".format(asset.dis_path)

    @override(Executor)
    def _open_workfiles(self, asset):
        self._open_dis_workfile(asset, fifo_mode=False)

    @override(Executor)
    def _open_workfiles_in_fifo_mode(self, asset):
        dis_p = multiprocessing.Process(target=self._open_dis_workfile,
                                        args=(asset, True))
        dis_p.start()
        self._wait_for_workfiles(asset)

    @override(Executor)
    def _open_procfiles(self, asset):
        self._open_dis_procfile(asset, fifo_mode=False)

    @override(Executor)
    def _open_procfiles_in_fifo_mode(self, asset):
        dis_p = multiprocessing.Process(target=self._open_dis_procfile,
                                        args=(asset, True))
        dis_p.start()
        self._wait_for_procfiles(asset)

    @override(Executor)
    def _close_workfiles(self, asset):
        self._close_dis_workfile(asset)

    @classmethod
    @override(Executor)
    def _close_procfiles(cls, asset):
        cls._close_dis_procfile(asset)

    @override(Executor)
    def _save_result(self, result):
        self.result_store.save(result)
        if self.save_workfiles:
            qw, qh = result.asset.quality_width_height
            self.result_store.save_workfile(result, result.asset.dis_workfile_path, f'_dis.{qw}x{qh}.{self._get_workfile_yuv_type(result.asset)}.yuv')
        return result



model_type = "5PL"
model_param_dict = {
    # ==== preprocess: normalize each feature ==== #
    # 'norm_type':'none', 
    'norm_type': 'clip_0to1', # rescale to within [0, 1]

    # ==== postprocess: clip final quality score ==== #
    'score_clip':[0.0, 100.0], # clip to within [0, 100]
}
